from wsln.lexer import Token, Phrase, TokenTag
from wsln.builder import W_SLN, SemanticLink, SemanticNode, LinkType
from wsln.reasoner import divide_link_community, ResolvedLink

def test_divide_link_community():
    link1 = SemanticLink(
                from_node=SemanticNode(Token("A", TokenTag.NOUN)),
                to_node=SemanticNode(Token("B", TokenTag.NOUN)),
                indicator_token=Phrase([
                    Token("as", TokenTag.PREPOSITION),
                    Token("well", TokenTag.ADV),
                    Token("as", TokenTag.PREPOSITION),
                ], TokenTag.ANY),
                link_type=LinkType.SEQUENTIAL_AND,
            )
    link2 = SemanticLink(
                from_node=SemanticNode(Token("B", TokenTag.NOUN)),
                to_node=SemanticNode(Token("good", TokenTag.ADJ)),
                indicator_token=Token("is", TokenTag.VERB),
                link_type=LinkType.ATTRIBUTE,
            )
    link3 = SemanticLink(
                from_node=SemanticNode(Token("A", TokenTag.NOUN)),
                to_node=SemanticNode(Token("B", TokenTag.NOUN)),
                indicator_token=Phrase([
                    Token("as", TokenTag.PREPOSITION),
                    Token("well", TokenTag.ADV),
                ], TokenTag.ANY),
                link_type=LinkType.SEQUENTIAL_AND,
            )
    link4 = SemanticLink(
                from_node=SemanticNode(Token("C", TokenTag.NOUN)),
                to_node=SemanticNode(Token("good", TokenTag.ADJ)),
                indicator_token=Token("is", TokenTag.VERB),
                link_type=LinkType.ATTRIBUTE,
            )
    link5 = SemanticLink(
                from_node=SemanticNode(Token("good", TokenTag.NOUN)),
                to_node=SemanticNode(Token("C", TokenTag.ADJ)),
                indicator_token=Token("is", TokenTag.VERB),
                link_type=LinkType.ATTRIBUTE,
            )
    

    test_cases = [
        (
            [
                {link1, link2},
                {link3},
                {link2, link4},
                {link5},
            ],
            [
                {link1, link2, link4},
                {link3},
                {link5},
            ],
        ),
        (
            [
                {link1, link2},
                {link3, link4},
                {link2, link4},
                {link5},
            ],
            [
                {link1, link2, link3, link4},
                {link5},
            ],
        ),
    ]

    for input_link_sets, expected_link_sets in test_cases:
        input_communities = [W_SLN(links=list(input_link_set)) for input_link_set in input_link_sets]
        communities = divide_link_community(input_communities)

        assert communities == [W_SLN(links=list(expected_link_set)) for expected_link_set in expected_link_sets]


def test_find_all_matched_links(reason_rules):
    # cause_effect_trans_rule, action_use_trans_rule = reason_rules

    car_node = SemanticNode(Token("car", TokenTag.NOUN))
    engine_node = SemanticNode(Token("engine", TokenTag.NOUN))
    gas_node = SemanticNode(Token("gas", TokenTag.NOUN))

    car_use_engine_link = SemanticLink(
        from_node=car_node,
        to_node=engine_node,
        link_type=LinkType.ACTION,
        indicator_token=(Token("use", TokenTag.VERB)),
    )

    engine_use_gas_link = SemanticLink(
        from_node=engine_node,
        to_node=gas_node,
        link_type=LinkType.ACTION,
        indicator_token=Phrase([Token("use", TokenTag.VERB), Token("yes", TokenTag.VERB)], TokenTag.VERB),
    )

    engine_use_no_gas_link = SemanticLink(
        from_node=engine_node,
        to_node=gas_node,
        link_type=LinkType.ACTION,
        indicator_token=Phrase([Token("use", TokenTag.VERB), Token("no", TokenTag.VERB)], TokenTag.VERB),
    )

    you_node = SemanticNode(Token("you", TokenTag.NOUN))
    leg_node = SemanticNode(Token("leg", TokenTag.NOUN))
    broken_node = SemanticNode(Token("broken", TokenTag.NOUN))

    you_cause_leg = SemanticLink(
        from_node=you_node,
        to_node=leg_node,
        link_type=LinkType.CAUSE_EFFECT,
        indicator_token=(Token("therefore", TokenTag.VERB)),
    )

    leg_cause_broken = SemanticLink(
        from_node=leg_node,
        to_node=broken_node,
        link_type=LinkType.CAUSE_EFFECT,
        indicator_token=(Token("therefore", TokenTag.VERB)),
    )

    test_cases = [
        (
            W_SLN(links=[
                car_use_engine_link,
                engine_use_gas_link,
            ]),
            [[
                (ResolvedLink(car_node, engine_node, LinkType.ACTION, ["use"]), car_use_engine_link),
                (ResolvedLink(engine_node, gas_node, LinkType.ACTION, ["use", "use yes"]), engine_use_gas_link),
            ]],
            [[
                ResolvedLink(car_node, gas_node, LinkType.MEANS, []),
            ]],
        ),
        (
            W_SLN(links=[
                you_cause_leg,
                leg_cause_broken,
                car_use_engine_link,
            ]),
            [[
                (ResolvedLink(you_node, leg_node, LinkType.CAUSE_EFFECT, []), you_cause_leg),
                (ResolvedLink(leg_node, broken_node, LinkType.CAUSE_EFFECT, []), leg_cause_broken),
            ]],
            [[
                ResolvedLink(you_node, broken_node, LinkType.CAUSE_EFFECT, [])
            ]],
        ),
        (
            W_SLN(links=[
                you_cause_leg,
                car_use_engine_link,
                engine_use_no_gas_link,
            ]),
            [],
            [],
        )
    ]

    for input_sln, expected_matched_links, expected_derived_links in test_cases:
        matched_links, derived_links = [], []
        for reason_rule in reason_rules:
            _matched, _derived = reason_rule.find_all_matched_links(input_sln)
            matched_links += _matched
            derived_links += _derived

        assert matched_links == expected_matched_links
        assert derived_links == expected_derived_links




from wsln.apps.word_scoring import smi

def test_smi():
    informative_tokens = [
        "georgia", "mayor", "assails", "governor's", "move", "to", "reopen", "beaches"
    ]
    detailed_tokens = "the tybee island city council voted to close the beaches on march 20 but georgia gov brian kemp issued a statewide shelter-in-place executive order which supersedes all local orders relating to coronavirus and also opened up the state's beaches".split(" ")
    returned = smi(informative_tokens, detailed_tokens, normalize=True)
    print(returned)



# add parent path to python path
import sys
import nltk
from pathlib import Path
sys.path.append(str(Path(__file__).absolute().parent.parent.parent))

from wsln.settings import version, pattern_handler
from wsln.lexer import Document, Sentence
from wsln.parser import Pattern, PatternSetHandler, LinkType, W_SLN
from wsln.apps.datasets import read_paper_dataset, read_cnn_corpus, read_legal_corpus
from wsln.apps.word_scoring import smi
from tqdm import tqdm
from typing import Dict, List, Set


recorder = {}

def print_msg(query_link, query_sentence, sum, message: str):
    global recorder
    # print(query_link)
    # print(query_sentence)
    # print(sum)
    # print(message)
    # print("=========")
    # print()

    if message not in recorder:
        recorder[message] = 0
    recorder[message] += 1


class SentenceScore:

    def __init__(self, index: int, sentence: Sentence, w_sln: W_SLN, score: float, selected: bool) -> None:
        self.index = index
        self.sentence = sentence
        self.w_sln = w_sln
        self.score = score
        self.selected = selected

        self.meaningful_words = self._meaningful_words()

    def _meaningful_words(self) -> Set[str]:
        word_set = set()
        for link in self.w_sln.links:
            word_set |= set(
                link.from_node.token.text.split(" ")
                + link.to_node.token.text.split(" ")
                + (link.indicator_node.token.text.split(" ") if link.link_type in [LinkType.ACTION, LinkType.ATTRIBUTE] else [])
            )
        return word_set - {''}

    def evaluate_score(self, rank_dict):
        def get_word_score(word):
            if word not in rank_dict:
                print(word, "not occurred")
            return rank_dict.get(word, 0)

        self.score = sum(get_word_score(word) for word in self.meaningful_words)



def simplified(summary_wsln, sentence):
    pass


def summarize(rank_dict: Dict[str, float], document: Document, d_r: float = 0.5, simplified_proportion: float = 0.3, metric="censmi", desired_length = 150) -> str:
    assert 0 <= d_r < 1
    assert 0 <= simplified_proportion < 1

    sentence_score_tuples = [
        SentenceScore(index, sentence, pattern_handler.match_sentence(sentence), 0., False)
        for index, sentence in enumerate(document.sentences)
    ]

    doc_wsln = W_SLN()
    for score_tuple in sentence_score_tuples:
        doc_wsln += score_tuple.w_sln

    summary_wsln = W_SLN()
    summary = ""
    summary_tuples = []
    # TODO: 改成预期单词数量
    # for _ in range(10):
    while len(summary.split(" ")) < desired_length:
        
        for sentence_score_tuple in sentence_score_tuples:
            score = 0
            for link in sentence_score_tuple.w_sln.links:
                for word in sentence_score_tuple.meaningful_words:
                    if word not in rank_dict:
                        print(f"{word} not in rank_dict")
                    score += rank_dict.get(word, 0)
            # TODO: 乘系数
            sentence_score_tuple.score = score

        sorted_sentences = sorted(
            filter(lambda x: x.selected is False, sentence_score_tuples),
            key = lambda x: -x.score
        )

        if not sorted_sentences:
            break

        selected_tuple = sorted_sentences[0]
        summary_tuples.append(selected_tuple)

        selected_tuple.selected = True

        redun_threshold = sorted(rank_dict.values())[int(len(rank_dict) * 0.2)]

        # 简化句子
        # for index, s_tuple in enumerate(summary_tuples):
        as_well_as = abs_red = attr_same = sit_same = pur = False
        red_nodes = []
        red_node_mapper = []
        for link in selected_tuple.w_sln.links:
            # 判断是否冗余
            red = False
            for token in (tokens := [link.from_node.token, link.indicator_node.token, link.to_node.token]):
                if all([rank_dict[word] < redun_threshold for word in token.text.split(" ") if word]):
                    red = True
                else:
                    red = False
                    break
            if red:
                print_msg(link, selected_tuple.sentence.text, summary, "redundant")
                red_nodes.extend([
                    link.from_node.token, link.indicator_node.token, link.to_node.token
                ])

            for s_link in summary_wsln.links:
                if link == s_link:
                    pass

                # for node in [link.from_node, link.indicator_node, link.to_node]:
                #     if isinstance(node.token, Phrase):
                #         for node.token.tokens in summary_wsln.nodes:
                #             print()
                #     pass

                # 判断abstract链接
                if link.link_type == LinkType.ABSTRACT and link.from_node == s_link.from_node and not abs_red:
                    # 那么把所有from_node给换成和这个from_node
                    # 还是看两个abstract链接中的from_node是否有交集？
                    # link.from_node
                    print_msg(link, selected_tuple.sentence.text, summary, "abstract redundant")
                    abs_red = True
                    red_node_mapper.append(
                        (link.from_node.token, link.to_node.token)
                    )
                    break
                    
                elif "as well as" in selected_tuple.sentence.text and not as_well_as:
                    print_msg(link, selected_tuple.sentence.text, summary, "as well as")
                    as_well_as = True
                elif link.link_type in [LinkType.ATTRIBUTE, LinkType.SIMILAR, LinkType.OWN] and link.to_node == s_link.to_node and link.from_node == s_link.from_node and link.link_type == s_link.link_type and not attr_same:
                    print_msg(link, selected_tuple.sentence.text, summary, "attr,sim,own all same")
                    attr_same = True
                    red_nodes.append(link.to_node.token)
                    break
                elif link.link_type in [LinkType.SITUATION, LinkType.CONDITION, LinkType.TEMPORAL, LinkType.CAUSE_EFFECT, LinkType.MEANS, LinkType.PURPOSE] and link.to_node == s_link.to_node and link.link_type == s_link.link_type and not sit_same:
                    print_msg(link, selected_tuple.sentence.text, summary, "sit, cond, temp, cause, means, pur to_node is same")
                    sit_same = True
                    red_nodes.append(link.to_node.token)
                    break
                # elif link.link_type in [LinkType.CONJ_AND, LinkType.CONJ_OR] and link.indicator_node.token.text == "as well as":
                #     print()
                elif link.link_type == LinkType.PURPOSE and link.indicator_node.token.text in ["aim to", "in order to", "want to", "would like to", "intend to"] and not pur:
                    print_msg(link, selected_tuple.sentence.text, summary, "pur is complex")
                    pur = True
                    red_nodes.append(link.to_node.token)
                    break
        
        if simplified_proportion:

            simplified = ""
            for token in selected_tuple.sentence.tokens:
                if token in red_nodes:
                    # print(token)
                    # print(token)
                    # simplified += token.text + " "
                    pass
                else:
                    for (ori_token, to_token) in red_node_mapper:
                        if token == ori_token:
                            # simplified += to_token.text + " "
                            break
                    else:
                        simplified += token.text + " "
            summary += simplified + ". "
        else:
            summary += selected_tuple.sentence.text

        summary_wsln += selected_tuple.w_sln
        # summary = " ".join([s.sentence.text for s in summary_tuples])

        if d_r:
            for word in selected_tuple.meaningful_words:
                rank_dict[word] *= d_r
                for c_word in doc_wsln.get_connected_tokens(word):
                    if c_word not in selected_tuple.meaningful_words:
                        rank_dict[word] /= d_r

    return summary


def rouge_score(reference, hypothesis):
    """
    reference: truth
    hypothesis: summary

    return: {
        "rouge-1": {"p": 1.0, "r": 1.0, "f": s}
        "rouge-2": ...,
        "rouge-l": ...,
    }
    """
    from rouge import Rouge
    rouge = Rouge()
    scores = rouge.get_scores(hypothesis, reference)
    return scores[0]


def statistics_link_in_summary(inputs):
    stat = {
        "summary": {},
        "text": {},
    }
    for abstract_sentences, text_sentences in inputs:
        for sentence in abstract_sentences:
            for link in pattern_handler.match_sentence(sentence).links:
                stat["summary"][link.link_type] = stat["summary"].get(link.link_type, 0) + 1

        for sentence in text_sentences:
            for link in pattern_handler.match_sentence(sentence).links:
                stat["text"][link.link_type] = stat["text"].get(link.link_type, 0) + 1

    print(stat)


def stat_main():
    statistics_link_in_summary((Document(abstract).sentences, Document(important + unimportant).sentences) for index, (abstract, important, unimportant) in enumerate(read_paper_dataset(), 1))
    statistics_link_in_summary((Document(abstract).sentences, Document(important + unimportant).sentences) for index, (abstract, important, unimportant) in enumerate(read_cnn_corpus(10000), 1))


if __name__ == "__main__":
    # stat_main()
    # exit()

    summarys = []
    abstracts = []

    def get_params():
        import itertools
        keys = ["d_r", "desired_length", "simplified_proportion"]
        values = [[0, 0.3], [120, 150], [0, 0.1, 0.2]]
        for value_list in itertools.product(*values):
            yield dict(zip(keys, value_list))

    score_recorder = []

    data = read_legal_corpus(500)
    # data = read_cnn_corpus(500)
    # data = list(read_paper_dataset())

    # for index, (abstract, important, unimportant) in enumerate(read_paper_dataset(), 1):
    # for index, (abstract, important, unimportant) in enumerate(read_cnn_corpus(500), 1):
    for index, (abstract, important, unimportant) in tqdm(enumerate(data, 1), total=len(data)):
        important = important.replace(".", ". ")
        unimportant = unimportant.replace(".", ". ")

        important_tokens = []
        # for sentence in nltk.sent_tokenize(important):
        for sentence in Document(important).sentences:
            important_tokens.extend([
                token.text for token in sentence.flatten()
            ])

        detailed_tokens = []
        # for sentence in nltk.sent_tokenize(unimportant):
        for sentence in Document(unimportant).sentences:
            detailed_tokens.extend([
                token.text for token in sentence.flatten()
            ])

        rank_dict = smi(important_tokens, detailed_tokens)
        wo_summary = summarize(rank_dict, Document(important) + Document(unimportant), d_r=0.7, simplified_proportion=0, desired_length=120)

        # summarys.append(nltk.sent_tokenize(summary))
        # abstracts.append(nltk.sent_tokenize(abstract))

        # wo_op_score = rouge_score(abstract, wo_summary)

        # rank_dict = smi(important_tokens, detailed_tokens)
        # w_summary = summarize(rank_dict, Document(important) + Document(unimportant), d_r=0.7, simplified_proportion=0.2, desired_length=120)

        # w_op_score = rouge_score(abstract, w_summary)

        # print(w_op_score['rouge-1']['f'] - wo_op_score['rouge-1']['f'])

        # print(w_summary)
        # print(wo_summary)
        # print("=======")

        # if w_op_score['rouge-1']['f'] - wo_op_score['rouge-1']['f'] > 0.15:
        #     print(w_summary)
        #     print(wo_summary)

        # summarys.append(nltk.sent_tokenize(summary))
        # abstracts.append(nltk.sent_tokenize(abstract))


    from metrics import rouge_perl
    # score = rouge_perl(abstracts, summarys, "test")
    # print(score)

    # score_recorder.append((params, score))

    print(recorder)
    print(score_recorder)

    exit()


from pathlib import Path
from tqdm import tqdm
from lxml import etree


def read_paper_dataset(limit=-1):
    # base_path = Path("/media/lee/辽东铁骑/数据集/acl2014/RST_summary/data/acl2014/")
    base_path = Path("/mnt/e/数据集/acl2014/RST_summary/data/acl2014/")

    paths = list((base_path / "abstract").glob("*.txt"))

    if limit == -1:
        limit = len(paths)

    for path in tqdm(paths[:limit], desc="paper-corpus"):
        yield read_paper(base_path, path.name)

def read_paper(base_path, file_name="147.P14-1087.xhtml.txt"):
    abstract = (base_path / "abstract" / file_name).read_text()
    introduction, *sections, conclusion = (base_path / "content" / file_name).read_text().split("\n")

    return abstract, introduction + conclusion, " ".join(sections)


def split_sentences(sentences, points):
        if len(points) <= 0: return sentences
        points.extend([0, 1])

        points = sorted(set(points))

        sentence_group = []
        sentence_len = len(sentences)
        for pre_point, after_point in zip(points[:-1], points[1:]):
            pre_point, after_point = map(int, 
                [sentence_len * pre_point, sentence_len * after_point])
            sentence_group.append(sentences[pre_point:after_point])

        return map(lambda lines: "".join(lines), sentence_group)


def read_cnn_corpus(limit=500):
    base_path = Path("/mnt/e/数据集/新闻/cnn/stories")
    items = []

    for file_path in tqdm(list(base_path.glob("*.story"))[:limit], desc="cnn-corpus"):
        item = read_cnn(file_path)
        items.append(item)
    
    return items

def read_cnn(file_path):
    content = file_path.read_text(encoding="utf-8")
    highlight_index = content.find("@highlight")
    story, highlights = content[:highlight_index], content[highlight_index:].split("@highlight")

    story_lines = [line for line in story.split("\n") if line]
    highlight_lines = [line for line in highlights if line]

    story_lines = [line[9:] if line.startswith("(CNN) -- ") else line for line in story_lines]
    highlight_lines = [line[9:] if line.startswith("(CNN) -- ") else line for line in highlight_lines]

    important, unimportant, important2 = list(split_sentences(story_lines, [0.5, 0.99]))
    important = important + important2

    return ". ".join(highlight_lines), important, unimportant


def read_legal_corpus(limit=3000):
    base_path = Path("/mnt/e/数据集/legal case/corpus/corpus")
    items = []

    for file_path in tqdm(list((base_path / "citations_summ").glob("*.xml"))[:limit], desc="legal-corpus"):
        items.append(read_legal(
            base_path / "citations_summ" / file_path.name,
            base_path / "fulltext" / file_path.name,
        ))
    
    return items


def read_legal(citation_path, fulltext_path):
    summ_tree = etree.HTML(citation_path.read_text(encoding="utf-8", errors="ignore"))
    full_tree = etree.HTML(fulltext_path.read_text(encoding="utf-8", errors="ignore"))

    summ_phrases = ". ".join(summ_tree.xpath("//citphrase//text()"))
    cite_phrases = ". ".join(full_tree.xpath("//catchphrase//text()"))
    sentences = full_tree.xpath("//sentence//text()")[:-2]  # Irrelevant sentence to text

    important, unimportant, important2 = list(split_sentences(sentences, [0.5, 0.99]))
    important = important + important2

    return f"{summ_phrases} {cite_phrases}", important, unimportant


import pyrouge
import os
import shutil
import codecs
import logging


def rouge_perl(ref, hyp, log_name):
    """Compute the Rouge score based ROUGE-Toolkits 1.5.5
    
    params:
        :ref [[sent1, sent2], [sent1, sent2]], 每个doc分句子输入
        :hyp [[sent1, sent2], [sent1, sent2]]
        :log_name 模型名称，在result/{log_name}目录下生成
    """

    assert len(ref) == len(hyp)
    log_dir = f"result/{log_name}/"

    if os.path.exists(log_dir):
        shutil.rmtree(log_dir)
    os.makedirs(log_dir)

    for i in range(len(ref)):
        with codecs.open(log_dir+"%06d_reference.txt" % i, 'w', 'utf-8') as f:
            #f.write(" ".join(ref[i]).replace(' ', ' ') + '\n')
            f.write("\n".join(ref[i]))
        with codecs.open(log_dir+"%06d_candidate.txt" % i, 'w', 'utf-8') as f:
            #f.write(" ".join(hyp[i]).replace(' ', ' ').replace('<unk>', 'UNK') + '\n')
            f.write("\n".join(hyp[i]))

    r = pyrouge.Rouge155()
    r.model_filename_pattern = '#ID#_reference.txt'
    r.system_filename_pattern = '(\d+)_candidate.txt'
    r.model_dir = log_dir   # reference dir
    r.system_dir = log_dir  # hypothesis dir
    logging.getLogger('global').setLevel(logging.WARNING)
    rouge_results = r.convert_and_evaluate(
    #    rouge_args='-a -2 -1 -c 95 -U -n 2 -w 1.2 -b 75'
    )
    scores = r.output_to_dict(rouge_results)

    return {
        "rouge-1": {
            "f": round(scores["rouge_1_f_score"] * 100, 2),
            "r": round(scores["rouge_1_recall"] * 100, 2),
            "p": round(scores["rouge_1_precision"] * 100, 2),
        },
        "rouge-2": {
            "f": round(scores["rouge_2_f_score"] * 100, 2),
            "r": round(scores["rouge_2_recall"] * 100, 2),
            "p": round(scores["rouge_2_precision"] * 100, 2),
        },
        "rouge-l": {
            "f": round(scores["rouge_l_f_score"] * 100, 2),
            "r": round(scores["rouge_l_recall"] * 100, 2),
            "p": round(scores["rouge_l_precision"] * 100, 2),
        },
    }

    recall = [round(scores["rouge_1_recall"] * 100, 2),
              round(scores["rouge_2_recall"] * 100, 2),
              round(scores["rouge_l_recall"] * 100, 2)]
    precision = [round(scores["rouge_1_precision"] * 100, 2),
                 round(scores["rouge_2_precision"] * 100, 2),
                 round(scores["rouge_l_precision"] * 100, 2)]
    f_score = [round(scores["rouge_1_f_score"] * 100, 2),
               round(scores["rouge_2_f_score"] * 100, 2),
               round(scores["rouge_l_f_score"] * 100, 2)]
    print("F_measure: %s Recall: %s Precision: %s\n"
              % (str(f_score), str(recall), str(precision)))
    
    with codecs.open(log_dir+"rougeScore", 'w+', 'utf-8') as f:
        f.write("F_measure: %s Recall: %s Precision: %s\n"
              % (str(f_score), str(recall), str(precision)))
    return f_score[:], recall[:], precision[:]


import re
from collections import namedtuple
from enum import Enum
from pathlib import Path
from typing import Union, Optional, List, Any, Tuple
from wsln.lexer import Sentence, Token, TokenTag, Phrase
from wsln.utils import load_json_to_meta

class Operand(Enum):
    
    NOUN = 1
    VERB = 2
    BE = 3
    ADJ = 4
    ADV = 5

    SLN = 6

    INDICATOR = 10


symbol_defs = {
    "N": (TokenTag.NOUN, []),
    "V": (TokenTag.VERB, []),
    "BE": (TokenTag.ANY, ["be", "is", "am", "are", "were", "was"]),
    "ADJ": (TokenTag.ADJ, []),
    "ADV": (TokenTag.ADV, []),
    "I": (TokenTag.ANY, []),
    "NUM": (TokenTag.NUMERAL, []),
    "S": (TokenTag.ANY, []),
}


class Symbol:

    def __init__(self, identifier, candidates : Optional[List[str]] = None) -> None:
        if not identifier:
            self.identifier = identifier
            self.token_tag = TokenTag.DEFAULT
            self.candidates = []
            return

        result = re.search(r"([A-Z]+)(\d*)", identifier)
        label, index = result.group(1), result.group(2)

        if not label in symbol_defs:
            raise Exception(f"unknown identifier `{identifier}`")

        token_tag, word_list = symbol_defs[label]
        candidates = (candidates if candidates else []) + word_list

        self.identifier = identifier
        self.label = label
        self.index = index
        self.token_tag = token_tag
        self.candidates = candidates

    def __eq__(self, other):
        if isinstance(other, Symbol):
            return self.identifier == other.identifier
        
        return False
    
    def __repr__(self):
        return f"Symbol({self.identifier}, {self.candidates})"

    def is_recursive(self):
        return self.label == "S"

    def match_tokens(self, tokens) -> int:
        if not self.candidates:
            return 1 if self.token_tag == tokens[0].token_tag else 0

        offsets = []
        prefix = ""

        # TODO: 可以提前终止，或者换Trie树
        for offset, token in enumerate(tokens, 1):
            prefix = ((prefix + " ") if prefix else "") + token.text

            for word in self.candidates:
                if word == prefix:
                    offsets.append(offset)
                    break

        return offsets[-1] if len(offsets) != 0 else 0


class LinkType(Enum):

    ACTION = 1
    ATTRIBUTE = 2
    NEGATIVE = 3

    SEQUENTIAL = 4
    CONJ_AND = 5
    CONJ_OR = 5.5
    SITUATION = 6
    CONDITION = 7
    CAUSE_EFFECT = 8
    ABSTRACT = 9
    PART_WHOLE = 9.5
    PURPOSE = 10
    SIMILAR = 11
    MEANS = 12
    OWN = 13

    TEMPORAL = 14

    NODE = 20

    # CO_OCCUR = 15


link_type_mapper = {
    "action": LinkType.ACTION,
    "attribute": LinkType.ATTRIBUTE,
    "negative": LinkType.NEGATIVE,
    "sequential": LinkType.SEQUENTIAL,
    "conj-and": LinkType.CONJ_AND,
    "conj-or": LinkType.CONJ_OR,
    "situation": LinkType.SITUATION,
    "condition": LinkType.CONDITION,
    "cause-effect": LinkType.CAUSE_EFFECT,
    "abstract": LinkType.ABSTRACT,
    "part-whole": LinkType.PART_WHOLE,
    "purpose": LinkType.PURPOSE,
    "similar": LinkType.SIMILAR,
    "means": LinkType.MEANS,
    "own": LinkType.OWN,
    "temporal": LinkType.TEMPORAL,
    "node": LinkType.NODE,
}


class SemanticNode:

    def __init__(self, content: Any) -> None:
        self.content = content

    @property
    def token(self) -> Optional[Token]:
        if isinstance(self.content, Token):
            return self.content
        return None

    @property
    def links(self) -> Optional[List[Token]]:
        if isinstance(self.content, list):
            return self.content
        return None

    @property
    def text(self) -> str:
        if self.token:
            return self.token.text
        if self.links:
            return ", ".join(str(link) for link in self.links)

    def __eq__(self, other) -> bool:
        if isinstance(other, SemanticNode):
            return self.content == other.content
        return False

    def __repr__(self) -> str:
        return f"SemanticNode({self.content})"

NULL_NODE = SemanticNode(Token(""))


class SemanticLink:

    def __init__(self, from_node: SemanticNode, indicator_node: SemanticNode, link_type: LinkType, to_node: SemanticNode):
        self.from_node = from_node
        self.indicator_node = indicator_node
        self.link_type = link_type
        self.to_node = to_node

    def __eq__(self, other) -> bool:
        if isinstance(other, SemanticLink):
            return (
                self.from_node == other.from_node
                and self.indicator_node == other.indicator_node
                and self.link_type == other.link_type
                and self.to_node == other.to_node
            )
        return False

    def __str__(self) -> str:
        return f"{self.from_node.text}-({self.indicator_node.text}, {self.link_type})->{self.to_node.text}"

    def __repr__(self) -> str:
        return f"SemanticLink({self.from_node}, {self.indicator_node}, {self.link_type}, {self.to_node})"


class W_SLN:

    def __init__(self, links: Optional[List[SemanticLink]] = None) -> None:
        self.links = links if links else []
    
    def __add__(self, other):
        self.links.extend(other.links)
        return self

    def __str__(self):
        return ", ".join(str(link) for link in self.links)
     
    def __repr__(self):
        return ", ".join(str(link) for link in self.links)
    
    def get_connected_tokens(self, text: str) -> List[Token]:
        tokens = []
        for link in self.links:
            if text in link.from_node.token.text or text in link.indicator_node.token.text or text in link.to_node.token.text:
                # TODO: token set
                tokens.extend(
                    link.from_node.token.flatten() + link.indicator_node.token.flatten() + link.to_node.token.flatten()
                )
        return [
            token
            for token in tokens if token.text != ""
        ]

    def search_incomplete_link(self,
            from_node: Optional[SemanticNode] = None,
            indicator_node: Optional[SemanticNode] = None,
            link_type: Optional[LinkType] = None,
            to_node: Optional[SemanticNode] = None,):

        def is_equal(node: SemanticNode, expected_node):
            return not expected_node or node == expected_node
        
        matched_links = []
        for link in self.links:
            if (is_equal(link.from_node, from_node)
                and is_equal(link.indicator_node, indicator_node)
                and is_equal(link.link_type, link_type)
                and is_equal(link.to_node, to_node)):
                
                matched_links.append(link)
        return matched_links

    def get_centroid(self, node: SemanticNode) -> int:
        centroid = 0
        for link in self.links:
            if node in [link.from_node, link.indicator_node, link.to_node]:
                centroid += 1
        return centroid


class MatchedPattern:

    def __init__(self, pattern, tokens: List[Union[List, Token]]):
        assert len(tokens) == len(pattern.symbols)
        self.pattern = pattern
        self.tokens = tokens

        self.symbol_identifiers = [symbol.identifier for symbol in self.pattern.symbols]
    
    @property
    def from_token(self):
        if not self.pattern.from_id:
            return Token("")
        return self.tokens[
            self.symbol_identifiers.index(self.pattern.from_id)
        ]

    @property
    def indicator_token(self):
        if not self.pattern.indicator:
            return Token("")
        return self.tokens[
            self.symbol_identifiers.index(self.pattern.indicator)
        ]

    @property
    def to_token(self) -> Optional[Token]:
        if self.pattern.to_id:
            return self.tokens[
                self.symbol_identifiers.index(self.pattern.to_id)
            ]
        return Token("")
        return 

    def _to_node(self, token: Token) -> SemanticNode:
        if isinstance(token, Token):
            return SemanticNode(token)
        if isinstance(token, MatchedPattern):
            return SemanticNode(token.to_link())
        if isinstance(token, list):
            return SemanticNode([
                mp.to_link()
                for mp in token
            ])
        return NULL_NODE

    def to_link(self) -> SemanticLink:
        return SemanticLink(
            self._to_node(self.from_token),
            self._to_node(self.indicator_token),
            self.pattern.link_type,
            self._to_node(self.to_token),
        )
    
    def __eq__(self, other):
        return self.pattern == other.pattern and self.tokens == other.tokens
    
    def __repr__(self):
        return f"MP({self.tokens})"


class Pattern:

    def __init__(self, pattern: str, from_id: str, indicator: str, link_type: Union[str, LinkType], to_id: str) -> None:
        self.symbols = [
            self._item_to_symbol(item)
            for item in re.findall(r"(\w+\(.*\)|\w+)", pattern)
        ]
        
        self.pattern = pattern
        self.from_id = from_id
        self.indicator = indicator
        self.to_id = to_id

        if isinstance(link_type, str):
            if link_type not in link_type_mapper:
                raise Exception(f"unknown link type {link_type}")
            self.link_type = link_type_mapper[link_type]
        elif isinstance(link_type, LinkType):
            self.link_type = link_type

    def _item_to_symbol(self, item: str) -> Symbol:
        """[summary]

        Args:
            item (str): like "V", "I1(like|love)"

        Returns:
            [type]: [description]
        """
        result = re.search(r"(\w+)(\(.+\))*", item)

        identifier = result.group(1)
        if (candidates := result.group(2)) is not None:
            candidates = candidates[1:-1].split("|")
        
        return Symbol(identifier, candidates)

    def match_tokens(self, tokens: List[Token]) -> List[MatchedPattern]:
        matched_patterns = []
        for sequence in self._match_tokens(self.symbols, tokens):
            symbols = [item[0] for item in sequence]
            tokens = [item[1] for item in sequence]

            if len(symbols) == len(self.symbols):
                matched_patterns.append(
                    MatchedPattern(self, tokens)
                )
        return matched_patterns

    def _find_all_matched_indexes(self, symbol, tokens) -> List[int]:
        indexes = []
        for token_index, token in enumerate(tokens):
            if (offset := symbol.match_tokens(tokens[token_index:])) != 0:
                indexes.append((token_index, offset))
        return indexes

    
    # TODO: 需要好好测试这个函数，太复杂了
    def _match_tokens(self, pend_symbols: List[Symbol], tokens: List[Token]) -> List[Tuple[Symbol, Token]]:
        matched_sequences = []
        first_symbol = pend_symbols[0]

        if first_symbol.is_recursive():
            if len(pend_symbols) == 1:
                candidate_sequences = [[(first_symbol, tokens)]]
            else:
                # 预留一个单词给match_offset
                candidate_sequences = []
                for match_offset, extra_offset in self._find_all_matched_indexes(pend_symbols[1], tokens[1:]):
                    matched_tokens = tokens[:match_offset + 1]
                    second_tokens = tokens[match_offset + 1: match_offset + 1 + extra_offset]

                    if extra_offset == 1:
                        second_tokens = second_tokens[0]
                    else:
                        second_tokens = Phrase(second_tokens)

                    candidate_sequences += [
                        [(first_symbol, matched_tokens), (pend_symbols[1], second_tokens)] + matched_sequence
                        for matched_sequence in self._match_tokens(pend_symbols[2:], tokens[match_offset + 1 + extra_offset:])
                    ]
            return candidate_sequences

        for token_index, token in enumerate(tokens):
            # TODO: 如果是phrase级别的匹配，怎么办。
            # 如果已经按照规则合并以后，那就是一个text object，句子与symbol的匹配应该按照text object来
            # Verb(does) ADV(not) 和 I(never|does not)之间的匹配比较麻烦
            # token phrase token 之间的匹配，以整个text object为单位
            if first_symbol.is_recursive():
                pass
            else:
                match_offset = first_symbol.match_tokens(tokens[token_index:])

                if match_offset == 0:
                    continue
                # TODO: 简化
                matched_tokens = tokens[token_index: token_index + match_offset]

                if match_offset == 1:
                    matched_tokens = matched_tokens[0]
                else:
                    matched_tokens = Phrase(tokens[token_index: token_index + match_offset])

                if len(pend_symbols) == 1:
                    candidate_sequences = [[(first_symbol, matched_tokens)]]
                else:
                    candidate_sequences = [
                        [(first_symbol, matched_tokens)] + matched_sequence
                        for matched_sequence in self._match_tokens(pend_symbols[1:], tokens[token_index + match_offset:])
                    ]

            matched_sequences += candidate_sequences
        
        return matched_sequences

    def match_tokens_recursive(self, tokens, patterns):
        """
        目前解决
        S1 I(if) S2的情况

        TODO: 还没有考虑到
        N S1 S2 N的情况
        """
        t_index = s_index = 0
        while t_index < len(tokens) and s_index < len(self.symbols):
            symbol = self.symbols[s_index]

            # if symbol.is_recursive():
            #     if len

            if not symbol.is_recursive() and (offset := self.symbols[s_index].match_tokens(tokens[t_index:])) != 0:
                t_index += offset
                s_index += 1

                index_recorder.append((t_index, offset))
                break
            else:
                t_index += 1

        if s_index != len(non_recursive_symbols):
            return []
        
        # for 


        for pattern in self.pattern:
            pattern.match_tokens(token[:index])

        token[:index], token[index: index + offset], token[index+offset+1]
        

    def __eq__(self, other):
        return (
            self.pattern == other.pattern
            and self.from_id == other.from_id
            and self.indicator == other.indicator
            and self.to_id == other.to_id
            and self.link_type == other.link_type
        )

    def __str__(self) -> str:
        from_symbol = ind_symbol = to_symbol = None
        for symbol in self.symbols:
            if symbol.identifier == self.from_id:
                from_symbol = symbol
            if symbol.identifier == self.indicator:
                ind_symbol = symbol
            if symbol.identifier == self.to_id:
                to_symbol = symbol
        from_symbol = from_symbol if from_symbol else "%%"
        ind_symbol = ind_symbol if ind_symbol else "%%"
        to_symbol = to_symbol if to_symbol else "%%"
        return f"{from_symbol} -- {ind_symbol} --> {to_symbol}"


def match_pattern_greedy(symbols, sentence):
    """Return ONE matched pattern by greedy algorithm, but return a list

    if the match failed: return None
    else: return a list to indicate the tuple of token and symbol

    https://leetcode-cn.com/problems/is-subsequence/solution/pan-duan-zi-xu-lie-by-leetcode-solution/

    Steps:
        The sentence is REQUIRED contain at most two noun (phrase).

    To quickly match, prefilter the pattern match by its node (pod) count, more than
    """
    matched_sequence = []

    symbol_index = token_index = 0
    while symbol_index < len(symbols) and token_index < len(sentence):
        if symbols[symbol_index] == sentence[token_index]:
            matched_sequence.append(
                (symbols[symbol_index], sentence[token_index])
            )
            symbol_index += 1
        
        token_index += 1

    if symbol_index < len(symbols):
        return []

    return [matched_sequence]


class PatternSetHandler:

    def __init__(self, patterns: Optional[List[Pattern]] = None):
        self.patterns = patterns if patterns else []

    def append_patterns(self, patterns: List[Pattern]) -> None:
        for pattern in patterns:
            self.patterns.append(pattern)
    
    def read_patterns_from_file(self, file_path: Union[str, Path]) -> None:
        patterns_json = load_json_to_meta(file_path)

        for link in patterns_json.links:
            link_type = link.type
            for pattern in link.patterns:
                self.patterns.append(Pattern(
                    pattern.pattern,
                    pattern.from_id,
                    pattern.indicator if hasattr(pattern, "indicator") else None,
                    link_type,
                    pattern.to_id,
                ))
    
    # TODO: patterns需要被排序
    # 拥有更specific的需要放在前面，拥有递归的需要放前面
    # TODO: 如果
    def match_tokens(self, tokens: List[Token]) -> List[MatchedPattern]:
        _ts = tokens
        tokens = [token for token in tokens]
        matched_patterns = []
        for pattern in self.patterns:
            _mpatterns = pattern.match_tokens(tokens)
            for _mpattern in _mpatterns:

                for index, (symbol, token) in enumerate(zip(_mpattern.pattern.symbols, _mpattern.tokens)):

                    if symbol.is_recursive():
                        for t in token:
                            tokens.remove(t)
                        _mpattern.tokens[index] = self.match_tokens(_mpattern.tokens[index])
                        # matched_patterns.append(_mpattern)
                        recursive = True
                    elif token not in [_mpattern.from_token, _mpattern.to_token] and token in tokens:
                        tokens.remove(token)
                        # matched_patterns.append(_mpattern)
                        appended = True
                    elif token not in [_mpattern.from_token, _mpattern.to_token] and isinstance(token, Phrase):
                        for t in token:
                            if t in tokens:
                                tokens.remove(t)
                #         matched_patterns.append(_mpattern)
                    elif token not in [_mpattern.from_token, _mpattern.to_token]:
                        if isinstance(token, Token) and token not in tokens:
                            break
                        # if isinstance(token, Phrase):
                        #     for t in token:
                        #         print(t, tokens, t in tokens)
                        #         if t not in tokens:
                        #             break
                        #     else:
                        #         continue
                        #     break

                # if recursive and not appended:
                #     matched_patterns.append(_mpattern)

                # if (indicator := _mpattern.indicator_token) in tokens:
                #     tokens.remove(indicator)
                #     matched_patterns.append(_mpattern)
                # 如果这个symbol是递归的，那么就移除这些tokens
                # 如何确保移除的就是正确的token
                # 每个token内部都有一个索引编号
                # 那么如何判断相等呢？
                # 内容上相等和hash remove之间的关系，需要验证一下
                else:
                    matched_patterns.append(_mpattern)


        return matched_patterns

    def _phrase_to_link(self, token: Token) -> List[SemanticLink]:
        if isinstance(token, Phrase) and len(token) == 2 and token[0].token_tag != token[1].token_tag:
            return [SemanticLink(
                SemanticNode(token), 
                SemanticNode(Token("")),
                LinkType.ABSTRACT,
                SemanticNode(token[0] if token[0].token_tag == token.token_tag else token[1])
            )] + self._phrase_to_link(token[0]) + self._phrase_to_link(token[1])
        return []

    def match_sentence(self, sentence: Sentence) -> W_SLN:
        links = []
        for tokens in sentence.noun_snippets:
            links.extend([
                matched_pattern.to_link()
                for matched_pattern in self.match_tokens(tokens)
            ])

        for token in sentence.tokens:
            links += self._phrase_to_link(token)

        return W_SLN(links)


from wsln.parser import SemanticLink, SemanticNode, LinkType
from wsln.settings import template_handler
from collections import namedtuple
from itertools import permutations


class PendLink:

    def __init__(self, from_node: str, to_node: str, link_type, link_words):
        self.from_node = from_node
        self.to_node = to_node
        self.link_type = link_type
        self.link_words = link_words

class ResolvedLink:

    def __init__(self, from_node: SemanticNode, to_node: SemanticNode, link_type: LinkType, link_words: list):
        self.from_node = from_node
        self.to_node = to_node
        self.link_type = link_type
        self.link_words = link_words

    def __eq__(self, other):
        if isinstance(other, SemanticLink):
            link = other

            return (True if self.link_words == [] else link.indicator_token.literal  in self.link_words) and (
                self.from_node == link.from_node and
                self.to_node == link.to_node and
                self.link_type == link.link_type
            )
        elif isinstance(other, ResolvedLink):
            return (
                self.from_node == other.from_node and 
                self.to_node == other.to_node and
                self.link_type == other.link_type and
                self.link_words == other.link_words
            )
        
        return False
    
    def __repr__(self):
        return f"{self.from_node}-{self.link_type}({self.link_words})->{self.to_node}"


class ReasonRule:

    def __init__(self, conditions, outs):
        self.conditions = conditions
        self.outs = outs

    @property
    def node_indicator_set(self):
        indicator_set = set()
        for condition_link in self.conditions:
            indicator_set.update([condition_link.from_node, condition_link.to_node])
        return indicator_set

    def find_all_matched_links(self, sln):
        """brute force to find all matched conditional links
        
        Find all unique nodes, and match them

        Args:
            link_set ([type]): [description]

        Returns:
            [type]: [description]
        """

        all_matched_links = []
        all_derived_links = []

        for nodes in permutations(sln.nodes, len(self.node_indicator_set)):
            indicator_node_mapper = {
                indicator: node for node, indicator in zip(nodes, self.node_indicator_set)
            }

            matched_links = []

            for resolved_link in [
                ResolvedLink(
                    from_node=indicator_node_mapper[condition_link.from_node],
                    to_node=indicator_node_mapper[condition_link.to_node],
                    link_type=condition_link.link_type,
                    link_words=condition_link.link_words,
                ) for condition_link in self.conditions
            ]:
                for link in sln.links:
                    if resolved_link == link:
                        matched_links.append((
                            resolved_link,
                            link,
                        ))
                        break
                else:
                    break
            else:
                # successful matched conditions
                all_matched_links.append(matched_links)
                all_derived_links.append([
                    ResolvedLink(
                        from_node=indicator_node_mapper[out_link.from_node],
                        to_node=indicator_node_mapper[out_link.to_node],
                        link_type=out_link.link_type,
                        link_words=out_link.link_words,
                    ) for out_link in self.outs
                ])
        
        return all_matched_links, all_derived_links


def read_reasoning_rules_from_file():
    pass

    return [ReasonRule()]


def divide_link_community(communities):
    """divide the links into reasoning communities, 
    the initial communities are the links extracted from different sentences.

    Args:
        communities (list): a list of SLN, where the links are all in a community
    """
    new_communities = []

    for community in communities:
        for new_community in new_communities:
            # no identical link

            if len(community.links | new_community.links) != len(community.links) + len(new_community.links):
                new_community.update(community)
                break
        else:
            new_communities.append(community)

    if len(communities) == len(new_communities):
        return new_communities

    return divide_link_community(new_communities)





from wsln.lexer import Token, Phrase, Sentence, TokenTag, merge_same_tags, merge_by_phrase_rule, PhraseRule, tokenize_text, segment_sentences


def test_phrase():
    test_cases = [
        (
            Phrase([
                Token("template", TokenTag.NOUN),
                Token("pattern", TokenTag.NOUN),
            ], TokenTag.NOUN),
            "template pattern",
        ),
    ]

    for input, expected in test_cases:
        assert input.text == expected


def test_merge_same_tags():

    test_cases = [
        (
            [
                Token("Jack", TokenTag.NOUN),
                Token("is", TokenTag.VERB),
                Token("a", TokenTag.ARTICLE),
                Token("link", TokenTag.NOUN),
                Token("boy", TokenTag.NOUN),
            ],
            [
                Token("Jack", TokenTag.NOUN),
                Token("is", TokenTag.VERB),
                Token("a", TokenTag.ARTICLE),
                Phrase([
                    Token("link", TokenTag.NOUN),
                    Token("boy", TokenTag.NOUN),
                ], TokenTag.NOUN)
            ],
        ),
        (
            [
                Token("Jack", TokenTag.NOUN),
                Token("is", TokenTag.VERB),
                Token("a", TokenTag.ARTICLE),
                Token("link", TokenTag.NOUN),
                Token("boy", TokenTag.NOUN),
                Token("girl", TokenTag.NOUN),
                Token("man", TokenTag.NOUN),
            ],
            [
                Token("Jack", TokenTag.NOUN),
                Token("is", TokenTag.VERB),
                Token("a", TokenTag.ARTICLE),
                Phrase([
                    Token("link", TokenTag.NOUN),
                    Token("boy", TokenTag.NOUN),
                    Token("girl", TokenTag.NOUN),
                    Token("man", TokenTag.NOUN),
                ], TokenTag.NOUN)
            ],
        )
    ]

    for input_tokens, expected_tokens in test_cases:
        merged_tokens = merge_same_tags(input_tokens)

        assert merged_tokens == expected_tokens


def test_merge_by_phrase_rule(phrase_rules):
    test_cases = [
        (
            [
                Token("Jack", TokenTag.NOUN),
                Token("is", TokenTag.VERB),
                Token("a", TokenTag.ARTICLE),
                Token("good", TokenTag.ADJ),
                Token("boy", TokenTag.NOUN),
            ],
            [
                Token("Jack", TokenTag.NOUN),
                Token("is", TokenTag.VERB),
                Token("a", TokenTag.ARTICLE),
                Phrase([
                    Token("good", TokenTag.ADJ),
                    Token("boy", TokenTag.NOUN),
                ], TokenTag.NOUN)
            ],
        ),
        (
            [
                Token("love", TokenTag.VERB),
                Token("baby", TokenTag.NOUN),
            ],
            [
                Token("love", TokenTag.VERB),
                Token("baby", TokenTag.NOUN),
            ],
        ),
        (
            [
                Token("Jack", TokenTag.NOUN),
                Token("is", TokenTag.VERB),
                Token("a", TokenTag.ARTICLE),
                Token("good", TokenTag.ADJ),
                Token("boy", TokenTag.NOUN),
                Token("ha", TokenTag.INTERJECTION),
            ],
            [
                Token("Jack", TokenTag.NOUN),
                Token("is", TokenTag.VERB),
                Token("a", TokenTag.ARTICLE),
                Phrase([
                    Token("good", TokenTag.ADJ),
                    Token("boy", TokenTag.NOUN),
                ], TokenTag.NOUN),
                Token("ha", TokenTag.INTERJECTION),
            ],
        ),
        (
            [
                Token("very", TokenTag.ADV),
                Token("good", TokenTag.ADJ),
                Token("boy", TokenTag.NOUN),
            ],
            [
                Phrase([
                    Phrase([
                        Token("very", TokenTag.ADV),
                        Token("good", TokenTag.ADJ),
                    ], TokenTag.ADJ),
                    Token("boy", TokenTag.NOUN),
                ], TokenTag.NOUN)
            ]
        ),
        (
            [
                Token("five", TokenTag.NUMERAL),
                Token("key", TokenTag.ADJ),
                Token("states", TokenTag.NOUN),
            ],
            [
                Phrase([
                    Token("five", TokenTag.NUMERAL),
                    Phrase([
                        Token("key", TokenTag.ADJ),
                        Token("states", TokenTag.NOUN),
                    ], TokenTag.NOUN),
                ], TokenTag.NOUN)
            ]
        )
    ]

    for input_tokens, expected_tokens in test_cases:
        merged_tokens = merge_by_phrase_rule(input_tokens, phrase_rules)

        assert merged_tokens == expected_tokens


def test_segment_sentences():
    test_cases = [
        (
            "This is sentence 1. This is sentence 2, phrase 1. This is sentence 3; This is sentence 4! What is sentence (5) ?Haha",
            [
                "This is sentence 1 .",
                "This is sentence 2 , phrase 1 .",
                "This is sentence 3 ; This is sentence 4 !",
                "What is sentence ( 5 ) ?",
                "Haha",
            ]
        )
    ]

    for input_text, expected_sentences in test_cases:
        sentences = segment_sentences(input_text)
        assert sentences == expected_sentences

def test_nltk_tokenizer():
    test_cases = [
        (
            "Jack is a good boy(five).",
            Sentence([
                Token("Jack", TokenTag.NOUN), 
                Token("is", TokenTag.VERB), 
                Token("a", TokenTag.ARTICLE), 
                Token("good", TokenTag.ADJ), 
                Token("boy", TokenTag.NOUN), 
                Token("(", TokenTag.PUNCTUATION), 
                Token("five", TokenTag.NUMERAL), 
                Token(")", TokenTag.PUNCTUATION), 
                Token(".", TokenTag.PUNCTUATION)
            ]),
        )
    ]

    for input_sentence, expected_tokens in test_cases:
        assert tokenize_text(input_sentence) == expected_tokens


def test_new_sentence():
    test_cases = [
        (
            "Jack is a good boy (five).",
            Sentence([
                Token("Jack", TokenTag.NOUN), 
                Token("is", TokenTag.VERB), 
                Token("a", TokenTag.ARTICLE), 
                Phrase([
                    Token("good", TokenTag.ADJ), 
                    Token("boy", TokenTag.NOUN), 
                ], TokenTag.NOUN),
                Token("(", TokenTag.PUNCTUATION),
                Token("five", TokenTag.NUMERAL), 
                Token(")", TokenTag.PUNCTUATION), 
                Token(".", TokenTag.PUNCTUATION)
            ]),
        ),
        (
            [
                Token("Jack", TokenTag.NOUN), 
                Token("is", TokenTag.VERB), 
                Token("a", TokenTag.ARTICLE), 
                Phrase([
                    Token("good", TokenTag.ADJ), 
                    Token("boy", TokenTag.NOUN), 
                ], TokenTag.NOUN),
                Token("(", TokenTag.PUNCTUATION), 
                Token("five", TokenTag.NUMERAL), 
                Token(")", TokenTag.PUNCTUATION), 
                Token(".", TokenTag.PUNCTUATION)
            ],
            Sentence("Jack is a good boy (five)."),
        ),
    ]

    for input, expected_sentence in test_cases:
        assert Sentence(input) == expected_sentence


def test_sentence_noun_snippets():
    test_cases = [
        (
            Sentence([
                Token("The", TokenTag.ARTICLE),
                Token("Pattern", TokenTag.NOUN),
                Token("consists", TokenTag.VERB),
                Token("of", TokenTag.PREPOSITION),
                Token("two", TokenTag.NUMERAL),
                Token("node", TokenTag.NOUN),
                Token("indicators", TokenTag.NOUN),
                Token("and", TokenTag.CONJUNCTION),
                Token("a", TokenTag.ARTICLE),
                Token("link", TokenTag.NOUN),
                Token("indicator", TokenTag.NOUN),
                Token("to", TokenTag.PREPOSITION),
                Token("extract", TokenTag.VERB),
                Token("links", TokenTag.NOUN),
                Token("ha", TokenTag.INTERJECTION)
            ]),
            [
                [
                    Token("The", TokenTag.ARTICLE),
                    Token("Pattern", TokenTag.NOUN),
                ],
                [
                    Token("Pattern", TokenTag.NOUN),
                    Token("consists", TokenTag.VERB),
                    Token("of", TokenTag.PREPOSITION),
                    Token("two", TokenTag.NUMERAL),
                    Token("node", TokenTag.NOUN)
                ],
                [
                    Token("node", TokenTag.NOUN),
                    Token("indicators", TokenTag.NOUN),
                ],
                [
                    Token("indicators", TokenTag.NOUN),
                    Token("and", TokenTag.CONJUNCTION),
                    Token("a", TokenTag.ARTICLE),
                    Token("link", TokenTag.NOUN),
                ],
                [
                    Token("link", TokenTag.NOUN),
                    Token("indicator", TokenTag.NOUN),
                ],
                [
                    Token("indicator", TokenTag.NOUN),
                    Token("to", TokenTag.PREPOSITION),
                    Token("extract", TokenTag.VERB),
                    Token("links", TokenTag.NOUN),
                ],
                [
                    Token("links", TokenTag.NOUN),
                    Token("ha", TokenTag.INTERJECTION),
                ]
            ]
        ),
        (
            Sentence([
                Token("Pattern", TokenTag.NOUN),
                Token("consists", TokenTag.VERB),
                Token("of", TokenTag.PREPOSITION),
                Token("two", TokenTag.NUMERAL),
                Token("node", TokenTag.NOUN),
            ]),
            [
                [
                    Token("Pattern", TokenTag.NOUN),
                    Token("consists", TokenTag.VERB),
                    Token("of", TokenTag.PREPOSITION),
                    Token("two", TokenTag.NUMERAL),
                    Token("node", TokenTag.NOUN)
                ],
            ]
        ),
    ]

    for input_sentence, expected_snippets in test_cases:
        assert input_sentence.noun_snippets == expected_snippets



import re
from collections import Counter, namedtuple
from enum import Enum
import string
import nltk
from typing import List, Union

symbols = r"！？｡＂＃＄％＆＇（）＊＋，－／：；＜＝＞＠［＼］＾＿｀｛｜｝～｟｠｢｣､、〃》「」『』【】〔〕〖〗〘〙〚〛〜〝〞〟〰〾〿–—‘'‛“”„‟…‧﹏."
symbols += string.punctuation + "\\"
symbols += "".join([chr(i) for i in range(945, 970)])
digits = "1234567890"


class TokenTag(Enum):
    DEFAULT = 0
    
    NOUN = 1
    VERB = 2
    ADV = 4
    ADJ = 5
    ARTICLE = 6
    CONJUNCTION = 7
    NUMERAL = 8

    PRONOUN = 9
    PREPOSITION = 10

    INDICATOR = 10

    PLACEHOLDER = 11

    INTERJECTION = 12

    WH_WORD = 13

    PUNCTUATION = 15

    ANY = 999


def nltk_pos_to_token_tag(pos_tag: str) -> TokenTag:
    mapper = {
        "CC": TokenTag.CONJUNCTION,
        "CD": TokenTag.NUMERAL,
        "DT": TokenTag.ARTICLE,
        "EX": TokenTag.PRONOUN,
        "FW": TokenTag.NOUN,
        "IN": TokenTag.PREPOSITION,
        "JJ": TokenTag.ADJ,
        "JJR": TokenTag.ADJ,
        "JJS": TokenTag.ADJ,
        "LS": TokenTag.ARTICLE,
        "MD": TokenTag.VERB,
        "NN": TokenTag.NOUN,
        "NNS": TokenTag.NOUN,
        "NNP": TokenTag.NOUN,
        "NNPS": TokenTag.NOUN,
        "PDT": TokenTag.ADJ,
        "POS": TokenTag.PLACEHOLDER,
        "PRP": TokenTag.NOUN,
        "PRP$": TokenTag.ADJ,
        "RB": TokenTag.ADV,
        "RBR": TokenTag.ADV,
        "RBS": TokenTag.ADV,
        "RP": TokenTag.PREPOSITION,
        "SYM": TokenTag.PLACEHOLDER,
        "TO": TokenTag.PREPOSITION,
        "UH": TokenTag.INTERJECTION,
        "VB": TokenTag.VERB,
        "VBD": TokenTag.VERB,
        "VBG": TokenTag.VERB,
        "VBN": TokenTag.VERB,
        "VBP": TokenTag.VERB,
        "VBZ": TokenTag.VERB,
        "WDT": TokenTag.WH_WORD,
        "WP": TokenTag.WH_WORD,
        "WP$": TokenTag.WH_WORD,
        "WRB": TokenTag.WH_WORD,
    }

    return mapper.get(pos_tag, TokenTag.PUNCTUATION)


class PhraseTag:

    ADJ_ADV = 1


class Tense:
    DEFAULT = 0

    PRESENT = 1


class Token:

    text: str
    token_tag: TokenTag
    
    def __init__(self, text, token_tag = TokenTag.DEFAULT):
        self.text = text
        self.token_tag = token_tag

    @property
    def is_ordinal(self):
        return (
            self.text in ["first", "second", "third"] or 
            ((self.token_tag in [TokenTag.ADJ, TokenTag.NUMERAL]) and self.text.endswith("th"))
        )

    def flatten(self):
        return [self]

    def __eq__(self, other):
        if isinstance(other, Token):
            return self.text == other.text and self.token_tag == other.token_tag
        return False
    
    def __repr__(self):
        return f"Token({self.text}, {self.token_tag})"

    def __hash__(self):
        return hash(repr(self))

NULL_TOKEN = Token("", TokenTag.ANY)


class Phrase(Token):

    def __init__(self, tokens: List[Token], token_tag: TokenTag = TokenTag.DEFAULT) -> None:
        self.tokens = tokens
        if not isinstance(token_tag, TokenTag):
            raise Exception(f"invalid token_tag: {token_tag}")
        self.token_tag = token_tag

    @property
    def text(self) -> str:
        return " ".join(to.text for to in self.tokens)

    def append(self, token: Token) ->  None:
        self.tokens.append(token)

    def flatten(self) -> List[Token]:
        flatten_tokens = []
        for token in self.tokens:
            if isinstance(token, Phrase):
                flatten_tokens.extend(token.flatten())
            else:
                flatten_tokens.append(token)
        return flatten_tokens

    def __len__(self):
        return len(self.tokens)

    def __getitem__(self, index):
        if isinstance(index, (int, slice)):
            return self.tokens[index]

        raise Exception("the index must be a int or slice")

    def __iter__(self):
        return iter(self.tokens)

    def __repr__(self):
        return f"Phrase({self.tokens}, {self.token_tag})"

    def __eq__(self, other):
        if isinstance(other, Phrase):
            if len(self) != len(other):
                return False
            
            return all(token == otoken for token, otoken in zip(self, other)) and self.token_tag == other.token_tag
        
        return False

    def __hash__(self):
        return hash(self.text)


def tokenize_text(text: str) -> List[Token]:
    import nltk
    words = nltk.word_tokenize(text)

    return Sentence([
        Token(word, nltk_pos_to_token_tag(pos_tag))
        for word, pos_tag in nltk.pos_tag(words)
    ])


def merge_same_tags(tokens: List[Token]) -> List[Token]:
    merged_tokens = []
    curr_token = tokens[0]

    for index in range(1, len(tokens)):
        next_token = tokens[index]

        if curr_token.token_tag == next_token.token_tag and curr_token.token_tag != TokenTag.PUNCTUATION:
            if isinstance(curr_token, Phrase):
                curr_token.append(next_token)
            else:
                curr_token = Phrase([curr_token, next_token], curr_token.token_tag)
        else:
            merged_tokens.append(curr_token)
            curr_token = next_token
    else:
        merged_tokens.append(curr_token)
    
    return merged_tokens


PhraseRule = namedtuple("PhraseRule", ["curr_token_tag", "next_token_tag", "token_tag"])

PHRASE_RULES = [
    PhraseRule(TokenTag.ADV, TokenTag.ADJ, TokenTag.ADJ),
    PhraseRule(TokenTag.ADJ, TokenTag.NOUN, TokenTag.NOUN),
    PhraseRule(TokenTag.NUMERAL, TokenTag.NOUN, TokenTag.NOUN), # six days
    PhraseRule(TokenTag.VERB, TokenTag.ADV, TokenTag.VERB),
    PhraseRule(TokenTag.ADV, TokenTag.VERB, TokenTag.VERB),
    PhraseRule(TokenTag.NOUN, TokenTag.ADV, TokenTag.NOUN),
]


def merge_by_phrase_rule(tokens: List[Token], phrase_rules) -> List[Token]:
    """TODO: nominate the core word

    Args:
        tokens (List[Token]): [description]
        phrase_rules ([type]): [description]

    Returns:
        [type]: [description]
    """
    merged_tokens = []

    index = 0
    while index < len(tokens) - 1:
        curr_token, next_token = tokens[index], tokens[index + 1]

        for phrase_rule in phrase_rules:
            if (curr_token.token_tag == phrase_rule.curr_token_tag and
                next_token.token_tag == phrase_rule.next_token_tag):
                merged_tokens.append(Phrase(
                    [curr_token, next_token], phrase_rule.token_tag
                ))
                index += 2
                break
        else:
            merged_tokens.append(curr_token)
            index += 1
    
    if index == len(tokens) - 1:
        merged_tokens.append(tokens[index])

    # until no new phrase can be merged
    if len(merged_tokens) != len(tokens):
        return merge_by_phrase_rule(merged_tokens, phrase_rules)

    return merged_tokens


class Sentence:

    def __init__(self, input: Union[List[Token], str]):
        if isinstance(input, str):
            text = input
            tokens = tokenize_text(text)
            tokens = merge_same_tags(tokens)
            tokens = merge_by_phrase_rule(tokens, PHRASE_RULES)
        elif isinstance(input, list):
            tokens = input
            # tokens = merge_same_tags(tokens)
            # tokens = merge_by_phrase_rule(tokens, PHRASE_RULES)
            text = ""
        else:
            raise Exception(f"unknown input {input}")

        self.text = text
        self.tokens = tokens

    def flatten(self) -> List[Token]:
        flatten_tokens = []
        for token in self.tokens:
            if isinstance(token, Phrase):
                flatten_tokens.extend(token.flatten())
            else:
                flatten_tokens.append(token)
        return flatten_tokens


    @property
    def noun_snippets(self):
        previous_noun_index = 0
        snippets = []

        for token_index, token in enumerate(self.tokens):
            if token.token_tag == TokenTag.NOUN and (token_index != 0 and token_index != len(self.tokens) - 1):
                snippets.append(self.tokens[previous_noun_index:token_index+1])
                previous_noun_index = token_index
        else:
            snippets.append(self.tokens[previous_noun_index:])

        return snippets
    
    def __getitem__(self, index):
        if isinstance(index, (int, slice)):
            return self.tokens[index]

        raise Exception("the index must be a int or slice")

    def __len__(self):
        return len(self.tokens)

    def __iter__(self):
        return iter(self.tokens)

    def __eq__(self, other):
        if isinstance(other, Sentence):
            return self.tokens == other.tokens
        
        return False

    def __next__(self):
        pass
    
    def __repr__(self):
        return "Sentence([" + ", ".join(str(token) for token in self.tokens) + "])"


def segment_sentences(text: str):
    import nltk

    for char in string.punctuation:
        if char in text:
            text = text.replace(char, f" {char} ")

    text = " ".join(text.split())

    return nltk.sent_tokenize(text)


class Document:

    def __init__(self, text: str, coref : bool = False):
        if coref:
            # TODO
            import spacy
            nlp = spacy.load('en_core_web_sm')
            import neuralcoref
            neuralcoref.add_to_pipe(nlp, greedyness=0.1)
            #greedyness默认是0.5，值越大提取到的coreference 越多，当然噪音越多。
            doc = nlp('My sister has a dog. She loves him.')
            print(doc._.has_coref)
            print(doc._.coref_clusters)
            raise NotImplementedError("coref")

        self.sentences = [
            Sentence(sentence) for sentence in nltk.sent_tokenize(text)
            # Sentence(sentence) for sentence in segment_sentences(text)
        ]

    def __add__(self, other):
        self.sentences.extend(other.sentences)
        return self


def segment_sentences(text) -> List[str]:
    import nltk

    for char in string.punctuation:
        if char in text:
            text = text.replace(char, f" {char} ")

    text = " ".join(text.split())

    return nltk.sent_tokenize(text)


# Process all the tokens about temporal information

from __future__ import annotations
from enum import Enum
from typing import List, Tuple, Optional, Dict, Union

class TimeGranularity(Enum):
    Second = 1
    Minute = 2
    Hour = 3
    Day = 4
    Week = 5
    WeekDay = 5.1
    Month = 6
    Quarter = 7
    Year = 8
    Decade = 9
    Century = 10
    Era = 11    # Dynamic time granularity


class BaseTimePoint(Enum):
    Now = 1
    Today = 2
    Monday = 3
    # FIRST_DAY = 4
    # SPRING = 6


class TimeFunction(Enum):
    Every = 1
    This = 2
    Last = 3
    Next = 4


class Tense(Enum):
    Future = 1
    Past = 2
    Now = 3


class TimePoint:

    def __init__(self, time_dict: Dict[TimeGranularity, Union[int]], tense: Tense = Tense.Now):
        self.time_dict = time_dict
        self.tense = tense

    def __str__(self) -> str:
        pass

    def __add__(self, other: TimePoint):
        time_dict = {}
        for gran in self.time_dict.keys() + other.time_dict.keys():
            pass

        self.time_dict
        return TimePoint()

    def __sub__(self, other: TimePoint):
        pass

    @property
    def is_monday(self):
        return self.time_dict.get(TimeGranularity.WeekDay) == 1


class SomeTime(TimePoint):

    def __init__(self):
        self.time_point = TimePoint({})

    def __add__(self, other: TimePoint):
        self.time_dict = self.time_dict + other


Now = SomeTime()


class TimeInterval:

    def __init__(self, interval: TimePoint) -> None:
        self.interval = interval


class TimeListInterval(TimeInterval):

    def __init__(self, times: List[TimePoint]):
        pass

    def __add__(self, other: TimePoint):
        pass

    def __sub__(self, other: TimePoint):
        pass


class StartEndInterval(TimeInterval):

    def __init__(self, start: TimePoint, end: TimePoint):
        self.start = start
        self.end = end

    @property
    def lasting_time(self):
        pass


AD = TimePoint({TimeGranularity.YEAR: 0})

# On monday, 指的是这周周一，还是普遍意义的周一？也许两者都有，取决于具体的语境
# 根据上一句中的时间可以推断出这句的on Monday含义
Monday = TimePoint({
    # TimeGranularity.Week: TimeFunction.Every,
    TimeGranularity.WeekDay: 1,
})
Tuesday = Monday + TimePoint({TimeGranularity.WeekDay: 1})
Wednesday = Tuesday + TimePoint({TimeGranularity.WeekDay: 1})
Thursday = Wednesday + TimePoint({TimeGranularity.WeekDay: 1})
Friday = Thursday + TimePoint({TimeGranularity.WeekDay: 1})
Saturday = Friday + TimePoint({TimeGranularity.WeekDay: 1})
Sunday = Saturday + TimePoint({TimeGranularity.WeekDay: 1})

January = TimePoint({
    TimeGranularity.Month: 1,
})
February = January + TimePoint({TimeGranularity.Month: 1})
March = February + TimePoint({TimeGranularity.Month: 1})
April = March + TimePoint({TimeGranularity.Month: 1})
May = April + TimePoint({TimeGranularity.Month: 1})
June = May + TimePoint({TimeGranularity.Month: 1})
July = June + TimePoint({TimeGranularity.Month: 1})
August = July + TimePoint({TimeGranularity.Month: 1})
September = August + TimePoint({TimeGranularity.Month: 1})
October = September + TimePoint({TimeGranularity.Month: 1})
November = October + TimePoint({TimeGranularity.Month: 1})
December = November + TimePoint({TimeGranularity.Month: 1})

Spring = TimeListInterval([February, March, April])
Summer = TimeListInterval([May, June, July])
Autumn = TimeListInterval([August, September, October])
Winter = TimeListInterval([November, December, January])

# 需要模糊的概念，正态分布函数
Morning = TimePoint({TimeGranularity.Day: TimeFunction.Every, TimeGranularity.Hour: [6, 7, 8, 9, 10, 11]})
Noon = TimePoint({TimeGranularity.Day: TimeFunction.Every, TimeGranularity.Hour: 12}),
Afternoon = TimeListInterval([
    TimePoint({TimeGranularity.Day: TimeFunction.Every, TimeGranularity.Hour: [13, 14, 15, 16, 17]}),
])
Evening = TimePoint({
    TimeGranularity.Day: TimeFunction.Every, TimeGranularity.Hour: [18, 19, 20],
})
Night = TimePoint({
    TimeGranularity.Day: TimeFunction.Every, TimeGranularity.Hour: [21, 22, 23],
})


def resolve_time(text_object):
    pass





# 时间的表达
# 1、今天
# 2、明天 = （今天，）
# 3、2000 = 元年 + （2000， 年） + （5，月） + （1，日）

# 时间是否重复，取决于起始时间是否是周期性的
# 3、一月份，重复性的时间，[StartYear, StartYear + （30，日）]
# 4、二月份 = 一月份 + （30, 日）
# 5、2000年一月份 = 元年 + （2000, 年） + [Start(Year), Start(Year) + （30，日）]


# add parent path to python path
import sys
from pathlib import Path
sys.path.append(str(Path(__file__).absolute().parent.parent))

from wsln.settings import version, pattern_handler
from wsln.lexer import Sentence

PROMPT = ">>> "

def main():
    print(f"W-SLN {version}")
    print('Type "help" for more information.\n')

    while (line := input(PROMPT)):
        sentence = Sentence(line.strip())
        print(sentence.tokens)

        w_sln = pattern_handler.match_sentence(sentence)
        for link in w_sln.links:
            print(link)

if __name__ == "__main__":
    main()


from wsln.temporal import (
    January, Saturday, TimeFunction, TimeInterval, TimePoint, Today, TimeGranularity, TimeListInterval
    Today,
    Monday, Tuesday, Wednesday, Week
    Weekday, Weekend
)

def test_resolve_atomic_time():
    cases = [
        ("today", Today),
        ("last week", Today - TimePoint({TimeGranularity.Week: -1})),
        ("every week", TimePoint({TimeGranularity.Week: TimeFunction.Every})),
        ("weekend", WeekEnd),
        ("weekday", WeekDay),
        ("the first month", January),
        ("January", January),
        ("9 am", TimePoint({TimeGranularity.Hour: 9})),
        ("9 pm", TimePoint({TimeGranularity.Hour: 21})),
        ("nine ten", TimePoint({TimeGranularity.Hour: 9, TimeGranularity.Mintue: 10})),
        ("nine ten", TimePoint({TimeGranularity.Hour: 9, TimeGranularity.Mintue: 10})),
        ("for two days", TimeInterval(Today, Today + {TimeGranularity.Day: 2})),
    ]

    for case in cases:
        pass




# def test_resolve_period_time():
#         (
#             Phrase([
#                 Token("good", TokenTag.ADJ),
#                 Token("morning", TokenTag.NOUN),
#             ], TokenTag.NOUN),
#             TimeInterval(BaseTime.TODAY, ),
#         ),

# def test_resolve_specific_time():
#     (
#         Phrase([
#             Token("good", TokenTag.ADJ),
#             Token("morning", TokenTag.NOUN),
#         ], TokenTag.NOUN),
#         TimeInterval(BaseTime.TODAY, ),
#     ),


# def test_resolve_compound_time():
#     cases = [
#         (
#             [
#                 Phrase([
#                     Token("two", TokenTag.ADJ),
#                     Token("weeks", TokenTag.NOUN),
#                 ]),
#                 Token("ago", TokenTag.ADV),
#             ],
#             TimePoint()
#         ),
#         (
#             Phrase()
#         )
#     ]


import sys
from pathlib import Path
sys.path.append(str(Path(__file__).absolute().parent.parent))

from wsln.lexer import new_sentence
from wsln.settings import template_handler

import bottle
from bottle import request, response, static_file
from bottle import get, post, route

@get("/index")
def index_page():
    return static_file("main_page.html", root="./static")

@get('/static/<path>')
def js(path):
    return static_file(path, root='./static')

@get("/links")
def extract_links():

    sentence = new_sentence(request.query["sentence"])

    print(sentence)

    links = template_handler.extract_links_from_sentence(sentence)

    node_set = set()
    for link in links:
        node_set.update({link.from_node.literal, link.to_node.literal})

    return {
        "nodes": list(node_set),
        "links": [{
                "from_node": link.from_node.literal,
                "to_node": link.to_node.literal,
                "link_word": link.indicator_token.literal,
                "link_type": str(link.link_type)[9:],
            } for link in links
        ]
    }

if __name__ == "__main__":
    bottle.run(server="gunicorn", host="0.0.0.0", port=8883)


from pathlib import Path
from wsln.parser import PatternSetHandler

version = "0.1"
patterns_file = Path(__file__).absolute().parent / "patterns/patterns.json"
pattern_handler = PatternSetHandler()
pattern_handler.read_patterns_from_file(patterns_file)


import setuptools

setuptools.setup(
    name = "wsln",
    version = "0.1.0",
    description = "Understand the Semantics of Text",
    author = "gajanlee",
    index_url = "https://github.com/gajanlee/W-SLN",
    author_email = "lee_jiazh@163.com",
    install_requires = ["nltk", "tqdm"],
    packages = setuptools.find_packages(),
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
    package_data = {
        "wsln": ["patterns/patterns.json"],
    },
)


import math
from collections import Counter
from functools import partial
from pathlib import Path
from tqdm import tqdm
from itertools import chain
from pathlib import Path

log = lambda x: 0 if x == 0 else math.log(x)



def tf(tokens):
    return Counter(tokens)

bbc_idf_dict = None
def init_idf_dict():
    global bbc_idf_dict
    from tokenizer import tokenize_sentences_and_words
    token_occurrence = {}

    import json
    config = json.load(open("corpus.json"))

    for file_path in tqdm(Path(config["bbc_corpus"]).glob("*.txt"), desc="bbc idf calculation"):
        text = file_path.read_text()
        for token in set(chain(*tokenize_sentences_and_words(text))):
            token_occurrence[token] = token_occurrence.get(token, 0) + 1

    for token in token_occurrence:
        token_occurrence[token] = 1 / token_occurrence[token]
    bbc_idf_dict = token_occurrence

def tf_idf(tokens):
    global bbc_idf_dict
    if not bbc_idf_dict:
        init_idf_dict()
    token_counter = Counter(tokens)
    return {
        token: token_counter.get(token) * bbc_idf_dict.get(token, 0) for token in tokens
    }

def smi(informative_tokens, detailed_tokens, lamda=0.3, normalize=True, debug=False):
    all_tokens = informative_tokens + detailed_tokens
    all_counter = Counter(all_tokens)
    informative_counter = Counter(informative_tokens)
    detailed_counter = Counter(detailed_tokens)

    def _smi(token):
        p_x_yi = informative_counter.get(token, 0) / len(all_tokens)
        p_x_yd = detailed_counter.get(token, 0) / len(all_tokens)

        p_x = all_counter.get(token) / len(all_tokens)
        p_yi = len(informative_tokens) / len(all_tokens)
        p_yd = len(detailed_tokens) / len(all_tokens)

        if p_x == 0 or p_yi == 0 or p_yd == 0:
            return 0

        return p_x_yi * log(p_x_yi / (p_x * p_yi)) - lamda * p_x_yd * log(p_x_yd / (p_x * p_yd))

    score_dict = {
        token: _smi(token) for token in all_counter
    }
    if normalize:
        if not score_dict:
            return score_dict

        max_score, min_score = max(score_dict.values()), min(score_dict.values())

        if max_score == min_score:
            return score_dict

        score_dict = {
            token: (value - min_score) / (max_score - min_score)
            for token, value in score_dict.items()
        }

    if debug:
        for word, _ in sorted(score_dict.items(), key=lambda x: x[1]):
            print(f"{word}\t{informative_counter.get(word, 0)}\t{detailed_counter.get(word, 0)}\t{all_counter.get(word, 0)}\t{score_dict[word]}")

        print(len(informative_tokens), len(detailed_tokens), len(all_tokens))

    return score_dict


infos = "p2 p2 p2 p3 p3 to reopen beaches".lower().split(" ")
details = "p2 p2 p2 p2 p2 p3 to close the beaches on March 20".lower().split(" ") + "But Georgia Gov Brian Kemp issued a statewide shelter-in-place executive order which super-sedes all local orders relating to coronavirus and also opened up the state beaches".lower().split(" ")


x = smi(
    infos,
    details,
    lamda=0.3, 
    normalize=False
)

p = smi(
    infos,
    details,
    lamda=-1, 
    normalize=False
)

info_counter = Counter(infos)
detail_counter = Counter(details)

for word, value in sorted(x.items(), key=lambda g: g[1], reverse=True):
    print(word, round(value, 3), round(p[word], 3), info_counter[word], detail_counter[word])

print(x)


import pytest
from pathlib import Path


@pytest.fixture(scope="session")
def phrase_rules():
    from wsln.lexer import PHRASE_RULES
    return PHRASE_RULES


@pytest.fixture(scope="session")
def pattern_set_handler_from_file():
    from wsln.parser import Pattern, PatternSetHandler, LinkType
    ph = PatternSetHandler()
    ph.read_patterns_from_file(Path(__file__).parent.parent / "patterns.json")

    return ph


@pytest.fixture(scope="session")
def pattern_set_handler():
    from wsln.parser import Pattern, PatternSetHandler, LinkType

    patterns = [
        Pattern(
            "N1 V N2",
            from_id="N1",
            indicator="V",
            to_id="N2",
            link_type=LinkType.ACTION,
        ),
        Pattern(
            "N1 V",
            from_id="N1",
            indicator="V",
            to_id="N1",
            link_type=LinkType.ACTION,
        ),
        Pattern(
            "N V ADJ",
            from_id="N",
            indicator="V",
            to_id="ADJ",
            link_type=LinkType.ATTRIBUTE,
        ),
        Pattern(
            "N1 I(and|as well as|besides) N2",
            from_id="N1",
            to_id="N2",
            indicator="I",
            link_type=LinkType.SEQUENTIAL_AND,
        ),
    ]

    return PatternSetHandler(patterns)


@pytest.fixture(scope="session")
def reason_rules():
    from wsln.builder import LinkType
    from wsln.reasoner import ReasonRule, PendLink

    return [
        ReasonRule(
            conditions=[
                PendLink("n1", "n2", LinkType.CAUSE_EFFECT, []),
                PendLink("n2", "n3", LinkType.CAUSE_EFFECT, []),
            ],
            outs=[
                PendLink("n1", "n3", LinkType.CAUSE_EFFECT, []),
            ],
        ),
        ReasonRule(
            conditions=[
                PendLink("n1", "n2", LinkType.ACTION, ["use"]),
                PendLink("n2", "n3", LinkType.ACTION, ["use", "use yes"]),
            ],
            outs=[
                PendLink("n1", "n3", LinkType.MEANS, []),
            ]
        ),
    ]



import sys
from pathlib import Path
sys.path.append(str(Path(__file__).absolute().parent.parent))

from wsln.lexer import Document
from wsln.parser import W_SLN

def view_sln_of_html(sln):
    pass

def view_sln_of_jpg(sln):
    pass

def view_sln_of_neo4j_statements(wsln: W_SLN):
    def _get_node_hash(text):
        return text.replace(" ", "_").replace("-", "_")
        
    def _get_link_hash(link):
        string = f"{link.link_name}_{_get_node_hash(link.literal)}"
        for appendix in link.appendix_links:
            string += f"_{appendix.link_name}_{_get_node_hash(appendix.literal)}"
        
        return string

    node_stats_set, link_stats_set = set(), set()
    for link in wsln.links:
        node_tag = "NODE"

        from_node_hash = _get_node_hash(link.from_node.text)
        to_node_hash = _get_node_hash(link.to_node.text)

        if from_node_hash:
            node_stats_set.add(
                f"CREATE ({from_node_hash}:{node_tag} {{name: '{link.from_node.text}'}})"
            )
        if to_node_hash:
            node_stats_set.add(
                f"CREATE ({to_node_hash}:{node_tag} {{name: '{link.to_node.text}'}})"
            )

        from_node_hash = from_node_hash or to_node_hash
        to_node_hash = to_node_hash or from_node_hash

        if len(link.indicator_node.text) <= 1:
            if to_node_hash in from_node_hash or from_node_hash in to_node_hash and from_node_hash != to_node_hash:
                link_text = "part_of"
            else:
                link_text = "co_occur"
        else:
            link_text = link.indicator_node.text
        link_stats_set.add(
            f"CREATE ({from_node_hash})-[:{_get_node_hash(link_text)}]->({to_node_hash})"
        )

    return list(node_stats_set) + list(link_stats_set)


if __name__ == "__main__":
    test_text = "Many approaches have been proposed previously to improve sentiment analysis on Twitter data. Even though some local contextual features could be help-ful to distinguish the two cases above, they still may not be enough to get the sentiment on the whole message correct. We established a state-of-the-art baseline that utilizes a variety of features, and built a topic-based sentiment mixture model with topic-specific Twitter data, all integrated in a semi-supervised training framework. We conduct pre-processing by removing stop words and some of the frequent words found in Twitter data. We intro-duce a topic-based mixture model for Twitter senti-ment. This motivates us to explore topic information explicitly in the task of sentiment analysis on Twitter data. This provides an opportunity for cross-domain topic identification when data from certain domain is more difficult to obtain than others. Lexi-cons with positive and negative words are important to sentiment classification."


    from wsln.settings import version, pattern_handler
    from wsln.lexer import Sentence

    wsln = W_SLN()
    
    for sentence in Document(test_text).sentences:
        wsln += pattern_handler.match_sentence(sentence)

    statements = view_sln_of_neo4j_statements(wsln)

    from pathlib import Path
    Path("./statements.txt").write_text("\n".join(statements))
    
    


from wsln.lexer import Token, TokenTag, Phrase, Sentence
from wsln.parser import Symbol, Pattern, MatchedPattern, LinkType, PatternSetHandler, SemanticLink, SemanticNode, W_SLN, NULL_NODE


def test_new_pattern():
    cases = [
        (
            (
                "N1 V N2",
                "N1",
                "V",
                "action",
                "N2",
            ),
            [
                Symbol("N1"),
                Symbol("V"),
                Symbol("N2"),
            ],
            LinkType.ACTION,
        ),
        (
            (
                "N1 BE V N2",
                "N1",
                "V",
                "attribute",
                "N2",
            ),
            [
                Symbol("N1"),
                Symbol("BE"),
                Symbol("V"),
                Symbol("N2")
            ],
            LinkType.ATTRIBUTE,
        ),
        (
            (
                "N1 I23(don't|do not|never) V N22",
                "N1",
                "I23",
                "negative",
                "N22",
            ),
            [
                Symbol("N1"),
                Symbol("I23", ["don't", "do not", "never"]),
                Symbol("V"),
                Symbol("N22"),
            ],
            LinkType.NEGATIVE,
        ),
        (
            (
                "N1 V",
                "N1",
                "V",
                "action",
                None,
            ),
            [
                Symbol("N1"),
                Symbol("V"),
            ],
            LinkType.ACTION,
        ),
        (
            (
                "S1 I(if) S2",
                "S1",
                "I",
                LinkType.CAUSE_EFFECT,
                "S2",
            ),
            [
                Symbol("S1"),
                Symbol("I", ["if"]),
                Symbol("S2"),
            ],
            LinkType.CAUSE_EFFECT,
        ),
    ]

    for input, expected_symbols, expected_link_type in cases:
        pattern = Pattern(*input)
        assert pattern.symbols == expected_symbols
        assert pattern.link_type == expected_link_type


def test_symbol_match_tokens():
    test_cases = [
        (
            Symbol("I", ["does", "does not", "does not like", "don't", "which", "never"]),
            [
                Token("I", TokenTag.NOUN),
                Phrase([Token("does", TokenTag.VERB), Token("not", TokenTag.ADV)], TokenTag.VERB),
                Token("like", TokenTag.VERB),
            ],
            0,
        ),
        (
            Symbol("I", ["does", "does not", "does not like", "don't", "which", "never"]),
            [
                Phrase([Token("does", TokenTag.VERB), Token("not", TokenTag.VERB)], TokenTag.VERB),
                Token("love", TokenTag.VERB),
            ],
            1,
        ),
        (
            Symbol("I", ["does", "does not", "does not like", "don't", "which", "never"]),
            [
                Phrase([Token("does", TokenTag.VERB), Token("not", TokenTag.VERB)], TokenTag.VERB),
                Token("like", TokenTag.VERB),
            ],
            2,
        ),
    ]

    for symbol, tokens, offset in test_cases:
        assert offset == symbol.match_tokens(tokens)


def test_pattern_private_match_tokens():
    test_cases = [
        (
            Pattern("N1 V N2", "N1", "V", "action", "N2"),
            [
                Token("Jack", TokenTag.NOUN),
                Token("is", TokenTag.VERB),
                Token("a", TokenTag.ARTICLE),
                Token("nice", TokenTag.ADJ),
            ],
            []
        ),
        (

        )
    ]


def test_pattern_match_tokens():
    test_cases = [
        (
            Pattern(
                "N1 BE V",
                "N1",
                "BE",
                "attribute",
                "V",
            ),
            [
                Token("Jack", TokenTag.NOUN),
                Token("is", TokenTag.VERB),
                Token("a", TokenTag.ARTICLE),
                Token("nice", TokenTag.ADJ),
            ],
            [],
        ),
        (
            Pattern(
                "N1 BE ADJ",
                "N1",
                "BE",
                "attribute",
                "ADJ",
            ),
            [
                Token("Jack", TokenTag.NOUN),
                Token("is", TokenTag.VERB),
                Token("a", TokenTag.ARTICLE),
                Token("good", TokenTag.ADJ),
            ],
            [[
                Token("Jack", TokenTag.NOUN),
                Token("is", TokenTag.VERB),
                Token("good", TokenTag.ADJ),
            ]],
        ),
        (
            Pattern(
                "N1 BE ADJ",
                "N1",
                "BE",
                "attribute",
                "ADJ",
            ),
            [
                Token("Jack", TokenTag.NOUN),
                Token("is", TokenTag.VERB),
                Token("a", TokenTag.ARTICLE),
                Phrase([
                    Token("good", TokenTag.ADJ),
                    Token("good", TokenTag.ADJ),
                ], TokenTag.ADJ),
            ],
            [[
                Token("Jack", TokenTag.NOUN),
                Token("is", TokenTag.VERB),
                Phrase([
                    Token("good", TokenTag.ADJ),
                    Token("good", TokenTag.ADJ),
                ], TokenTag.ADJ),
            ]],
        ),
        (
            Pattern(
                "N1 V N2",
                "N1",
                "V",
                "action",
                "N2",
            ),
            [
                Token("Jack", TokenTag.NOUN),
                Token("rob", TokenTag.VERB),
                Token("a", TokenTag.ARTICLE),
                Token("car", TokenTag.NOUN),
            ],
            [[
                Token("Jack", TokenTag.NOUN),
                Token("rob", TokenTag.VERB),
                Token("car", TokenTag.NOUN),
            ]]
        ),
        (
            Pattern(
                "N1 V N2",
                "N1",
                "V",
                "action",
                "N2",
            ),
            [
                Token("Jack", TokenTag.NOUN),
                Token("rob", TokenTag.VERB),
                Token("a", TokenTag.ARTICLE),
                Token("benz", TokenTag.NOUN),
                Token("car", TokenTag.NOUN),
            ],
            [
                [
                    Token("Jack", TokenTag.NOUN),
                    Token("rob", TokenTag.VERB),
                    Token("benz", TokenTag.NOUN),
                ],
                [
                    Token("Jack", TokenTag.NOUN),
                    Token("rob", TokenTag.VERB),
                    Token("car", TokenTag.NOUN),
                ]
            ]
        ),
        (
            Pattern(
                "N1 V",
                "N1",
                "V",
                "action",
                None,
            ),
            [
                Token("Jack", TokenTag.NOUN),
                Token("rob", TokenTag.VERB),
                Token("a", TokenTag.ARTICLE),
                Token("benz", TokenTag.NOUN),
                Token("car", TokenTag.NOUN),
            ],
            [[
                Token("Jack", TokenTag.NOUN),
                Token("rob", TokenTag.VERB),
            ]]
        ),
        (
            Pattern(
                "N1 V N2",
                "N1",
                "V",
                "action",
                "N2",
            ),
            [
                Token("Jack", TokenTag.NOUN),
                Token("rob", TokenTag.VERB),
                Token("and", TokenTag.CONJUNCTION),
                Token("burn", TokenTag.VERB),
                Token("car", TokenTag.NOUN),
            ],
            [
                [
                    Token("Jack", TokenTag.NOUN),
                    Token("rob", TokenTag.VERB),
                    Token("car", TokenTag.NOUN),
                ],
                [
                    Token("Jack", TokenTag.NOUN),
                    Token("burn", TokenTag.VERB),
                    Token("car", TokenTag.NOUN),
                ], 
            ]
        ),
        (
            Pattern(
                "N1 I2(don't|do not|does not|never) N2",
                "N1",
                "I2",
                "negative",
                "N2",
            ),
            [
                Token("Jack", TokenTag.NOUN),
                Token("does", TokenTag.VERB),
                Token("not", TokenTag.ADV),
                Token("football", TokenTag.NOUN),
            ],
            [
                [
                    Token("Jack", TokenTag.NOUN),
                    Phrase([
                        Token("does", TokenTag.VERB),
                        Token("not", TokenTag.ADV),
                    ]),
                    Token("football", TokenTag.NOUN),
                ],
            ],
        ),
        (
            Pattern(
                "N1 I2(don't|do not|does not|never) N2",
                "N1",
                "I2",
                "negative",
                "N2",
            ),
            [
                Token("Jack", TokenTag.NOUN),
                Token("never", TokenTag.ADV),
                Token("play", TokenTag.VERB),
                Token("football", TokenTag.NOUN),
            ],
            [[
                Token("Jack", TokenTag.NOUN),
                Token("never", TokenTag.ADV),
                Token("football", TokenTag.NOUN),
            ]],
        ),
        (
            Pattern(
                "N1 I2(don't|do not|does not|never) V N2",
                "N1",
                "I2",
                "negative",
                "N2",
            ),
            [
                Token("Jack", TokenTag.NOUN),
                Token("never", TokenTag.ADV),
                Token("play", TokenTag.VERB),
                Token("football", TokenTag.NOUN),
            ],
            [[
                Token("Jack", TokenTag.NOUN),
                Token("never", TokenTag.ADV),
                Token("play", TokenTag.VERB),
                Token("football", TokenTag.NOUN),
            ]],
        ),
        (
            Pattern(
                "N1 BE I(kind of) N2",
                "N1",
                "I",
                "abstract",
                "N2",
            ),
            [
                Token("Apple", TokenTag.NOUN),
                Token("is", TokenTag.VERB),
                Token("a", TokenTag.ARTICLE),
                Token("kind", TokenTag.NOUN),
                Token("of", TokenTag.PREPOSITION),
                Token("fruit", TokenTag.NOUN),
            ],
            [[
                Token("Apple", TokenTag.NOUN),
                Token("is", TokenTag.VERB),
                Phrase([
                    Token("kind", TokenTag.NOUN),
                    Token("of", TokenTag.PREPOSITION),
                ]),
                Token("fruit", TokenTag.NOUN),
            ]],
        ),
        (
            Pattern("S1 I(if|because) S2", "S1", "I", "cause-effect", "S2"),
            [
                Token("I", TokenTag.NOUN),
                Token("could", TokenTag.VERB),
                Token("do", TokenTag.VERB),
                Token("if", TokenTag.ADV),
                Token("you", TokenTag.NOUN),
                Token("like", TokenTag.VERB),
            ],
            [[
                [
                    Token("I", TokenTag.NOUN),
                    Token("could", TokenTag.VERB),
                    Token("do", TokenTag.VERB)
                ],
                Token("if", TokenTag.ADV),
                [
                    Token("you", TokenTag.NOUN),
                    Token("like", TokenTag.VERB),
                ],
            ]]
        ),
        (
            Pattern("S1 I(if|because|because of) S2", "S1", "I", "cause-effect", "S2"),
            [
                Token("I", TokenTag.NOUN),
                Token("could", TokenTag.VERB),
                Token("do", TokenTag.VERB),
                Token("if", TokenTag.ADV),
                Token("you", TokenTag.NOUN),
                Token("like", TokenTag.VERB),
                Token("because", TokenTag.ADV),
                Token("they", TokenTag.NOUN),
                Token("hate", TokenTag.VERB),
            ],
            [
                [
                    [
                        Token("I", TokenTag.NOUN),
                        Token("could", TokenTag.VERB),
                        Token("do", TokenTag.VERB),
                    ],
                    Token("if", TokenTag.ADV),
                    [
                        Token("you", TokenTag.NOUN),
                        Token("like", TokenTag.VERB),
                        Token("because", TokenTag.ADV),
                        Token("they", TokenTag.NOUN),
                        Token("hate", TokenTag.VERB),
                    ]
                ],
                [
                    [
                        Token("I", TokenTag.NOUN),
                        Token("could", TokenTag.VERB),
                        Token("do", TokenTag.VERB),
                        Token("if", TokenTag.ADV),
                        Token("you", TokenTag.NOUN),
                        Token("like", TokenTag.VERB),
                    ],
                    Token("because", TokenTag.ADV),
                    [
                        Token("they", TokenTag.NOUN),
                        Token("hate", TokenTag.VERB),
                    ]
                ],
            ]
        ),
        (
            Pattern("S1 I(if|because|because of) S2", "S1", "I", "cause-effect", "S2"),
            [
                Token("I", TokenTag.NOUN),
                Token("could", TokenTag.VERB),
                Token("do", TokenTag.VERB),
                Token("if", TokenTag.ADV),
                Token("you", TokenTag.NOUN),
                Token("like", TokenTag.VERB),
                Token("because", TokenTag.ADV),
                Token("of", TokenTag.PREPOSITION),
                Token("they", TokenTag.NOUN),
                Token("hate", TokenTag.VERB),
            ],
            [
                [
                    [
                        Token("I", TokenTag.NOUN),
                        Token("could", TokenTag.VERB),
                        Token("do", TokenTag.VERB),
                    ],
                    Token("if", TokenTag.ADV),
                    [
                        Token("you", TokenTag.NOUN),
                        Token("like", TokenTag.VERB),
                        Token("because", TokenTag.ADV),
                        Token("of", TokenTag.PREPOSITION),
                        Token("they", TokenTag.NOUN),
                        Token("hate", TokenTag.VERB),
                    ]
                ],
                [
                    [
                        Token("I", TokenTag.NOUN),
                        Token("could", TokenTag.VERB),
                        Token("do", TokenTag.VERB),
                        Token("if", TokenTag.ADV),
                        Token("you", TokenTag.NOUN),
                        Token("like", TokenTag.VERB),
                    ],
                    Phrase([
                        Token("because", TokenTag.ADV),
                        Token("of", TokenTag.PREPOSITION),
                    ]),
                    [
                        Token("they", TokenTag.NOUN),
                        Token("hate", TokenTag.VERB),
                    ]
                ],
            ]
        )
    ]

    for pattern, tokens, expected in test_cases:
        actual = pattern.match_tokens(tokens)
        assert actual == [
            MatchedPattern(pattern, expected_tokens)
            for expected_tokens in expected
        ]


def test_match_pattern_recursive_link():
    pattern_action1 = Pattern(
        "N1 V N2",
        from_id="N1",
        indicator="V",
        to_id="N2",
        link_type=LinkType.ACTION,
    )
    pattern_action2 = Pattern(
        "N1 V",
        from_id="N1",
        indicator="V",
        to_id=None,
        link_type=LinkType.ACTION,
    )
    pattern_if = Pattern(
        "S1 I(if) S2",
        "S1",
        "I",
        "cause-effect",
        "S2",
    )
    test_cases = [
        (
            [pattern_if, pattern_action1, pattern_action2],
            [
                Token("I", TokenTag.NOUN),
                Token("could", TokenTag.VERB),
                Token("do", TokenTag.VERB),
                Token("if", TokenTag.ADV),
                Token("you", TokenTag.NOUN),
                Token("like", TokenTag.VERB),
            ],
            [
                SemanticLink(
                    from_node=SemanticNode([SemanticLink(
                        SemanticNode(Token("I", TokenTag.NOUN)),
                        SemanticNode(
                            Token("could", TokenTag.VERB),
                        ),
                        LinkType.ACTION,
                        NULL_NODE,
                    ), SemanticLink(
                        SemanticNode(Token("I", TokenTag.NOUN)),
                        SemanticNode(Token("do", TokenTag.VERB),),
                        LinkType.ACTION,
                        NULL_NODE,
                    )]),
                    indicator_node=SemanticNode(Token("if", TokenTag.ADV)),
                    link_type=LinkType.CAUSE_EFFECT,
                    to_node=SemanticNode([SemanticLink(
                        SemanticNode(Token("you", TokenTag.NOUN)),
                        SemanticNode(Token("like", TokenTag.VERB)),
                        LinkType.ACTION,
                        NULL_NODE,
                    )]),
                )
            ],
        ),
    ]

    for patterns, tokens, expected_links in test_cases:
        ph = PatternSetHandler(patterns)
        matched_patterns = ph.match_tokens(tokens)
        assert [mp.to_link() for mp in matched_patterns] == expected_links


def test_pattern_set_handler_match_tokens():
    pattern_action1 = Pattern(
        "N1 V N2",
        from_id="N1",
        indicator="V",
        to_id="N2",
        link_type=LinkType.ACTION,
    )
    pattern_action2 = Pattern(
        "N1 V",
        from_id="N1",
        indicator="V",
        to_id="N1",
        link_type=LinkType.ACTION,
    )
    pattern_attr = Pattern(
        "N V ADJ",
        from_id="N",
        indicator="V",
        to_id="ADJ",
        link_type=LinkType.ATTRIBUTE,
    )
    pattern_conj = Pattern(
        "N1 I(and|as well as|besides) N2",
        from_id="N1",
        to_id="N2",
        indicator="I",
        link_type=LinkType.CONJ_AND,
    )
    pattern_abs = Pattern(
        "N1 BE I(kind of|class of) N2",
        "N1", "I", LinkType.ABSTRACT, "N2",
    )
    pattern_node = Pattern(
        "N",
        None, "N", LinkType.NODE, None,
    )
    pattern_if = Pattern("S1 I(if) S2", "S1", "I", "action", "S2")

    pattern_set_handler = PatternSetHandler([
        pattern_abs, pattern_if, pattern_action1, pattern_action2, pattern_attr, pattern_conj, pattern_node 
    ])

    test_cases = [
        (
            [
                Token("I", TokenTag.NOUN),
                Token("don't", TokenTag.ADV),
                Token("like", TokenTag.VERB),
                Token("and", TokenTag.CONJUNCTION),
                Token("love", TokenTag.VERB),
                Token("the", TokenTag.ARTICLE),
                Token("guy", TokenTag.NOUN),
                Token("writing", TokenTag.VERB),
            ],
            [
                MatchedPattern(pattern_action1, [
                    Token("I", TokenTag.NOUN),
                    Token("like", TokenTag.VERB),
                    Token("guy", TokenTag.NOUN),
                ]),
                MatchedPattern(pattern_action1, [
                    Token("I", TokenTag.NOUN),
                    Token("love", TokenTag.VERB),
                    Token("guy", TokenTag.NOUN),
                ]),
                MatchedPattern(pattern_action2, [
                    Token("I", TokenTag.NOUN),
                    Token("writing", TokenTag.VERB),
                ]),
                MatchedPattern(pattern_conj, [
                    Token("I", TokenTag.NOUN),
                    Token("and", TokenTag.CONJUNCTION),
                    Token("guy", TokenTag.NOUN),
                ]),
                MatchedPattern(pattern_node, [
                    Token("I", TokenTag.NOUN),
                ]),
                MatchedPattern(pattern_node, [
                    Token("guy", TokenTag.NOUN),
                ]),
            ],
        ),
        (
            [
                Token("I", TokenTag.NOUN),
                Token("could", TokenTag.VERB),
                Token("do", TokenTag.VERB),
                Token("if", TokenTag.ADV),
                Token("you", TokenTag.NOUN),
                Token("like", TokenTag.VERB),
            ],
            [
                MatchedPattern(pattern_if, [
                    [
                        MatchedPattern(pattern_action2, [
                        Token("I", TokenTag.NOUN),
                        Token("could", TokenTag.VERB),
                        ]),
                        MatchedPattern(pattern_action2, [
                            Token("I", TokenTag.NOUN),
                            Token("do", TokenTag.VERB),
                        ]),
                        MatchedPattern(pattern_node, [
                            Token("I", TokenTag.NOUN),
                        ]),
                    ],
                    Token("if", TokenTag.ADV),
                    [
                        MatchedPattern(pattern_action2, [
                            Token("you", TokenTag.NOUN),
                            Token("like", TokenTag.VERB),
                        ]),
                        MatchedPattern(pattern_node, [
                            Token("you", TokenTag.NOUN),
                        ]),
                    ],
                ])
            ],
        ),
        (
            [
                Token("Apple", TokenTag.NOUN),
                Token("is", TokenTag.VERB),
                Token("kind", TokenTag.NOUN),
                Token("of", TokenTag.PREPOSITION),
                Token("fruit", TokenTag.NOUN),
                Token(".", TokenTag.PUNCTUATION),
            ],
            [
                MatchedPattern(pattern_abs, [
                    Token("Apple", TokenTag.NOUN),
                    Token("is", TokenTag.VERB),
                    Phrase([
                        Token("kind", TokenTag.NOUN),
                        Token("of", TokenTag.PREPOSITION),
                    ]),
                    Token("fruit", TokenTag.NOUN),
                ]),
                MatchedPattern(pattern_node, [
                    Token("Apple", TokenTag.NOUN),
                ]),
                MatchedPattern(pattern_node, [
                    Token("fruit", TokenTag.NOUN),
                ]),
            ],
        ),
        (
            [
                Phrase([
                    Token("Red", TokenTag.ADJ),
                    Token("Apple", TokenTag.NOUN),
                ], TokenTag.NOUN),
                Token("is", TokenTag.VERB),
                Phrase([
                    Token("kind", TokenTag.NOUN),
                    Token("of", TokenTag.PREPOSITION),
                ]),
                Token("fruit", TokenTag.NOUN),
                Token(".", TokenTag.PUNCTUATION),
            ],
            [
                MatchedPattern(pattern_abs, [
                    Phrase([
                        Token("Red", TokenTag.ADJ),
                        Token("Apple", TokenTag.NOUN),
                    ], TokenTag.NOUN),
                    Token("is", TokenTag.VERB),
                    Phrase([
                        Token("kind", TokenTag.NOUN),
                        Token("of", TokenTag.PREPOSITION),
                    ]),
                    Token("fruit", TokenTag.NOUN),
                ]),
                MatchedPattern(pattern_node, [
                    Phrase([
                        Token("Red", TokenTag.ADJ),
                        Token("Apple", TokenTag.NOUN),
                    ], TokenTag.NOUN),
                ]),
                MatchedPattern(pattern_node, [
                    Token("fruit", TokenTag.NOUN),
                ]),
            ],
        ),
        (
            [Token("Apple", TokenTag.NOUN)],
            [MatchedPattern(pattern_node, [Token("Apple", TokenTag.NOUN)])]
        ),
        (
            [
                Token("Apple", TokenTag.NOUN),
                Token("orange", TokenTag.NOUN)
            ],
            [
                MatchedPattern(pattern_node, [Token("Apple", TokenTag.NOUN)]),
                MatchedPattern(pattern_node, [Token("orange", TokenTag.NOUN)]),
            ]
        ),
        (
            [
                Token("Apple", TokenTag.NOUN),
                Token("lele", TokenTag.PREPOSITION),
                Token("orange", TokenTag.NOUN),
            ],
            [
                MatchedPattern(pattern_node, [Token("Apple", TokenTag.NOUN)]),
                MatchedPattern(pattern_node, [Token("orange", TokenTag.NOUN)]),
            ]
        ),
    ]

    for tokens, expected_matched_patterns in test_cases:
        before_tokens = [token for token in tokens]
        assert pattern_set_handler.match_tokens(tokens) == expected_matched_patterns
        assert tokens == before_tokens


def test_phrase_to_link():
    token_five = Token("five", TokenTag.NUMERAL)
    token_key = Token("key", TokenTag.ADJ)
    token_states = Token("states", TokenTag.NOUN)
    token_key_states = Phrase([token_key, token_states], TokenTag.NOUN)
    token_five_key_states = Phrase([token_five, token_key_states], TokenTag.NOUN)

    test_cases = [
        (
            token_five_key_states,
            [
                (token_five_key_states, token_key_states),
                (token_key_states, token_states)
            ]
        )
    ]

    ph = PatternSetHandler()
    for phrase, expected in test_cases:
        actual = ph._phrase_to_link(phrase)
        expected = [
            SemanticLink(SemanticNode(expected_from), SemanticNode(Token("")), LinkType.ABSTRACT, SemanticNode(expected_to))
            for expected_from, expected_to in expected
        ]

        assert actual == expected


def test_handler_match_sentence():
    ph = PatternSetHandler([
        Pattern("N1 BE I(kind of) N2", "N1", "I", "abstract", "N2"),
        Pattern("N1 V N2", "N1", "V", "action", "N2"),
        Pattern("N1 V", "N1", "V", "action", None),
        Pattern("S1 I(if) S2", "S1", "I", "condition", "S2"),
    ])

    test_cases = [
        (
            Sentence("Apple is a kind of fruit."),
            W_SLN([
                SemanticLink(
                    SemanticNode(Token("Apple", TokenTag.NOUN)),
                    SemanticNode(Phrase([
                        Token("kind", TokenTag.NOUN),
                        Token("of", TokenTag.PREPOSITION),
                    ])),
                    LinkType.ABSTRACT,
                    SemanticNode(Token("fruit", TokenTag.NOUN)),
                )
            ]),
        )
    ]

    for input_sentence, expected_wsln in test_cases:
        print(input_sentence)
        # assert ph.match_sentence(input_sentence) == expected_wsln


def test_pattern_set_handler_from_file_match_sentence(pattern_set_handler_from_file):
    assert len(pattern_set_handler_from_file.patterns) > 0

    test_cases = [
        (
            Sentence("If he wins five key states, Republican candidate Mitt Romney will be elected President in 2008."),
            [
                ("he", "If", LinkType.CONDITION, "he"),
                ("he", "wins", LinkType.ACTION, "five key states"),
                ("five key states", ",", LinkType.SEQUENTIAL, "Republican candidate Mitt Romney"),
                ("Republican candidate Mitt Romney", "will be elected", LinkType.ACTION, "President"),
                ("President", "in", LinkType.SITUATION, "2008"),
                ("five key states", "", LinkType.ABSTRACT, "key states"),
                ("key states", "", LinkType.ABSTRACT, "states"),
                ("Republican candidate Mitt Romney", "", LinkType.ABSTRACT, "candidate Mitt Romney"),
            ]
        ),
        # (
        #     Sentence("I will come with you if he likes."),
        #     []
        # ),
        # (
        #     Sentence("Apple is a kind of fruit."),
        #     [],
        # ),
    ]

    for sentence, expected_links in test_cases:
        w_sln = pattern_set_handler_from_file.match_sentence(sentence)

        assert len(w_sln.links) == len(expected_links)

        for link, (from_text, indicator_text, link_type, to_text) in zip(w_sln.links, expected_links):
            assert (
                link.from_node.token.text == from_text
                and link.indicator_node.token.text == indicator_text
                and link.link_type == link_type
                and link.to_node.token.text == to_text
            )


def test_wsln_search_incomplete_link():
    link_I_love_your_dog = SemanticLink(
        SemanticNode(Token("I")),
        SemanticNode(Token("love")),
        LinkType.ACTION,
        SemanticNode(Phrase([Token("your"), Token("dog")])),
    )
    link_I_am_good = SemanticLink(
        SemanticNode(Token("I")),
        SemanticNode(Token("am")),
        LinkType.ATTRIBUTE,
        SemanticNode(Token("good"))
    )
    link_I_and_dog = SemanticLink(
        SemanticNode(Token("I")),
        SemanticNode(Token("and")),
        LinkType.CONJ_AND,
        SemanticNode(Token("dog")),
    )
    link_dog_barking = SemanticLink(
        SemanticNode(Token("dog")),
        SemanticNode(Token("barking")),
        LinkType.ACTION,
        SemanticNode(Token("dog")),
    )

    test_cases = [
        (
            [
                link_I_love_your_dog,
                link_I_am_good,
                link_I_and_dog,
                link_dog_barking,
            ],
            [
                (
                    {
                        "from_node": SemanticNode(Token("I")),
                    },
                    [
                        link_I_love_your_dog,
                        link_I_am_good,
                        link_I_and_dog,
                    ]
                ),
                (
                    {
                        "link_type": LinkType.ACTION,
                    },
                    [
                        link_I_love_your_dog,
                        link_dog_barking,
                    ]
                ),
                (
                    {
                        "from_node": SemanticNode(Token("I")),
                        "to_node": SemanticNode(Token("dog")),
                    },
                    [
                        link_I_and_dog,
                    ]
                ),
                (
                    {
                        "to_node": SemanticNode(Phrase([Token("your"), Token("dog")])),
                    },
                    [
                        link_I_love_your_dog,
                    ]
                )
            ]
        ),
    ]

    for input_links, conditions in test_cases:
        w_sln = W_SLN(input_links) 
        for condition, expected_links in conditions:
            assert w_sln.search_incomplete_link(**condition) == expected_links